import functools
import logging
from typing import Any

from opentelemetry.trace import Status, StatusCode

from ...contexts import GenerationContext
from ...sdk import VerseSDK
from .utils import (
    format_messages_for_tracing,
    get_completion_choices,
    get_finish_reason,
    get_request_parameters,
    get_stream_text,
    get_usage_from_response,
)


def configure_anthropic(sdk: VerseSDK):
    """Configure Anthropic to use the Verse SDK."""
    try:
        import anthropic
    except ImportError:
        logging.warning("Anthropic SDK not installed, skipping integration")
        return

    original_create = anthropic.resources.messages.Messages.create
    original_create_async = anthropic.resources.messages.AsyncMessages.create
    original_stream = anthropic.resources.messages.Messages.stream

    @functools.wraps(original_create)
    def create_with_tracing(self, *args, **kwargs):
        """Wrap synchronous messages.create with tracing."""
        model = kwargs.get("model", "unknown")
        messages = kwargs.get("messages", [])
        stream = kwargs.get("stream", False)

        generation_ctx = sdk.create_generation(
            name=f"Anthropic: {'Chat' if messages else 'Completion'}",
            model=model,
            usage=get_request_parameters(kwargs),
        )

        generation_ctx.provider("anthropic").operation(
            "chat" if messages else "text_completion"
        ).request_params(get_request_parameters(kwargs))

        if messages:
            formatted_messages = format_messages_for_tracing(messages)
            generation_ctx.messages(formatted_messages)

        try:
            response = original_create(self, *args, **kwargs)

            if stream:
                return _wrap_stream_response(response, generation_ctx, sdk)

            _process_response(response, generation_ctx)
            generation_ctx.span.end()
            return response

        except Exception as e:
            generation_ctx.error(e)
            generation_ctx.span.end()
            raise

    @functools.wraps(original_create_async)
    async def create_async_with_tracing(self, *args, **kwargs):
        """Wrap asynchronous messages.create with tracing."""
        model = kwargs.get("model", "unknown")
        messages = kwargs.get("messages", [])
        stream = kwargs.get("stream", False)

        generation_ctx = sdk.create_generation(
            name=f"Anthropic: {'Chat' if messages else 'Completion'}",
            model=model,
            usage=get_request_parameters(kwargs),
        )

        generation_ctx.provider("anthropic").operation(
            "chat" if messages else "text_completion"
        ).request_params(get_request_parameters(kwargs))

        if messages:
            formatted_messages = format_messages_for_tracing(messages)
            generation_ctx.messages(formatted_messages)

        try:
            response = await original_create_async(self, *args, **kwargs)

            if stream:
                return _wrap_async_stream_response(response, generation_ctx, sdk)

            _process_response(response, generation_ctx)
            generation_ctx.span.end()
            return response

        except Exception as e:
            generation_ctx.error(e)
            generation_ctx.span.end()
            raise

    @functools.wraps(original_stream)
    def stream_with_tracing(self, *args, **kwargs):
        """Wrap synchronous messages.stream with tracing."""
        model = kwargs.get("model", "unknown")
        messages = kwargs.get("messages", [])

        generation_ctx = sdk.create_generation(
            name="Anthropic: Chat",
            model=model,
            usage=get_request_parameters(kwargs),
        )

        generation_ctx.provider("anthropic").operation("chat").request_params(
            get_request_parameters(kwargs)
        )

        if messages:
            formatted_messages = format_messages_for_tracing(messages)
            generation_ctx.messages(formatted_messages)

        try:
            stream = original_stream(self, *args, **kwargs)
            return _wrap_stream_context_manager(stream, generation_ctx, sdk)
        except Exception as e:
            generation_ctx.error(e)
            generation_ctx.span.end()
            raise

    class TracedStreamContextManager:
        """Wrapper for MessageStream context manager that traces events."""

        def __init__(self, stream_manager, generation_ctx: GenerationContext, sdk: VerseSDK):
            self._stream_manager = stream_manager
            self._generation_ctx = generation_ctx
            self._sdk = sdk
            self._finalized = False
            self._inner_stream = None

        def __enter__(self):
            self._inner_stream = self._stream_manager.__enter__()
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            try:
                return self._stream_manager.__exit__(exc_type, exc_val, exc_tb)
            finally:
                self._finalize()

        def __iter__(self):
            """Iterate over stream events, emit deltas, and finalize span."""
            if self._inner_stream is None:
                raise RuntimeError(
                    "TracedStreamContextManager must be entered with 'with' before iteration"
                )
            try:
                for event in self._inner_stream:
                    if (
                        event.type == "content_block_delta"
                        and hasattr(event.delta, "text")
                        and event.delta.text is not None
                    ):
                        self._sdk.event(
                            "gen_ai.stream.delta",
                            metadata={"content": event.delta.text[:512]},
                            level="info",
                            span=self._generation_ctx.span,
                        )
                    yield event
            except Exception as e:
                self._generation_ctx.error(e)
                raise
            finally:
                self._finalize()

        def get_final_message(self):
            """Get the final message from the stream."""
            if self._inner_stream is None:
                raise RuntimeError(
                    "TracedStreamContextManager must be entered with 'with' before calling get_final_message()"
                )
            return self._inner_stream.get_final_message()

        def _finalize(self):
            """Finalize the span using get_final_message if available."""
            if self._finalized:
                return
            self._finalized = True

            try:
                if self._inner_stream:
                    final_message = self._inner_stream.get_final_message()
                    if final_message:
                        _process_response(final_message, self._generation_ctx)
            except Exception as e:
                logging.warning("Error finalizing Anthropic stream", exc_info=e)
            finally:
                self._generation_ctx.span.end()

    def _wrap_stream_context_manager(stream, generation_ctx: GenerationContext, sdk: VerseSDK):
        """Wrap streaming context manager with tracing proxy."""
        return TracedStreamContextManager(stream, generation_ctx, sdk)

    anthropic.resources.messages.Messages.create = create_with_tracing
    anthropic.resources.messages.AsyncMessages.create = create_async_with_tracing
    anthropic.resources.messages.Messages.stream = stream_with_tracing


def _process_response(response: Any, generation_ctx: GenerationContext):
    """Process non-streaming response and add attributes to span."""
    try:
        if hasattr(response, "model"):
            generation_ctx.model_used(response.model)

        if hasattr(response, "id"):
            generation_ctx.response_id(response.id)

        finish_reason = get_finish_reason(response)
        if finish_reason:
            generation_ctx.reason(finish_reason)

        choices = get_completion_choices(response)
        if choices:
            generation_ctx.completions(choices)

        usage = get_usage_from_response(response)
        if usage:
            generation_ctx.usage(usage)

        generation_ctx.status(Status(StatusCode.OK))

    except Exception as e:
        logging.warning("Error processing Anthropic response", exc_info=e)


class TracedMessageStream:
    """Proxy wrapper for MessageStream that ensures span finalization."""

    def __init__(self, stream, generation_ctx: GenerationContext, sdk: VerseSDK):
        self._stream = stream
        self._generation_ctx = generation_ctx
        self._sdk = sdk
        self._finalized = False

    def __getattr__(self, name):
        """Delegate attribute access to the underlying stream."""
        return getattr(self._stream, name)

    def __iter__(self):
        """Iterate over stream chunks, emit deltas, and finalize span."""
        try:
            for chunk in self._stream:
                text = get_stream_text(chunk)
                if text:
                    self._sdk.event(
                        "gen_ai.stream.delta",
                        metadata={"content": text[:512]},
                        level="info",
                        span=self._generation_ctx.span,
                    )
                yield chunk
        except Exception as e:
            self._generation_ctx.error(e)
            raise
        finally:
            self._finalize()

    def __enter__(self):
        """Enter context manager."""
        self._stream.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context manager and finalize span."""
        try:
            return self._stream.__exit__(exc_type, exc_val, exc_tb)
        finally:
            self._finalize()

    def _finalize(self):
        """Finalize the span using get_final_message if available."""
        if self._finalized:
            return
        self._finalized = True

        try:
            if hasattr(self._stream, "get_final_message"):
                final_message = self._stream.get_final_message()
                _process_response(final_message, self._generation_ctx)
            elif hasattr(self._stream, "response"):
                _process_response(self._stream.response, self._generation_ctx)
        except Exception as e:
            logging.warning("Error finalizing Anthropic stream", exc_info=e)
        finally:
            self._generation_ctx.span.end()


class TracedAsyncMessageStream:
    """Proxy wrapper for AsyncMessageStream that ensures span finalization."""

    def __init__(self, stream, generation_ctx: GenerationContext, sdk: VerseSDK):
        self._stream = stream
        self._generation_ctx = generation_ctx
        self._sdk = sdk
        self._finalized = False

    def __getattr__(self, name):
        """Delegate attribute access to the underlying stream."""
        return getattr(self._stream, name)

    async def __aiter__(self):
        """Iterate over stream chunks, emit deltas, and finalize span."""
        try:
            async for chunk in self._stream:
                text = get_stream_text(chunk)
                if text:
                    self._sdk.event(
                        "gen_ai.stream.delta",
                        metadata={"content": text[:512]},
                        level="info",
                        span=self._generation_ctx.span,
                    )
                yield chunk
        except Exception as e:
            self._generation_ctx.error(e)
            raise
        finally:
            await self._finalize()

    async def __aenter__(self):
        """Enter async context manager."""
        await self._stream.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit async context manager and finalize span."""
        try:
            return await self._stream.__aexit__(exc_type, exc_val, exc_tb)
        finally:
            await self._finalize()

    async def _finalize(self):
        """Finalize the span using get_final_message if available."""
        if self._finalized:
            return
        self._finalized = True

        try:
            if hasattr(self._stream, "get_final_message"):
                final_message = await self._stream.get_final_message()
                _process_response(final_message, self._generation_ctx)
            elif hasattr(self._stream, "response"):
                _process_response(self._stream.response, self._generation_ctx)
        except Exception as e:
            logging.warning("Error finalizing Anthropic async stream", exc_info=e)
        finally:
            self._generation_ctx.span.end()


def _wrap_stream_response(stream, generation_ctx: GenerationContext, sdk: VerseSDK):
    """Wrap streaming response with tracing proxy."""
    return TracedMessageStream(stream, generation_ctx, sdk)


def _wrap_async_stream_response(
    stream, generation_ctx: GenerationContext, sdk: VerseSDK
):
    """Wrap async streaming response with tracing proxy."""
    return TracedAsyncMessageStream(stream, generation_ctx, sdk)



import asyncio
import types
from typing import Any
from unittest.mock import Mock

from verse_sdk.integrations.pydantic_ai.patch import patch_pydantic_ai


class DummySpan:
    def __init__(self):
        self.ended = False

    def set_attribute(self, key: str, value: Any):
        pass

    def end(self):
        self.ended = True


class DummyGenerationContext:
    def __init__(self):
        self.span = DummySpan()
        self.captured_messages = None
        self.captured_completions = None
        self.captured_reason = None

    def messages(self, messages):
        self.captured_messages = messages

    def completions(self, completions):
        self.captured_completions = completions

    def reason(self, reason):
        self.captured_reason = reason

    def provider(self, provider):
        self.captured_provider = provider
        return self

    def operation(self, operation):
        self.captured_operation = operation
        return self


def make_part(part_type: str, **attrs):
    part_cls = type(part_type, (), {})
    part = part_cls()
    for key, value in attrs.items():
        setattr(part, key, value)
    return part


def test_patch_pydantic_ai_records_messages_and_completions(monkeypatch):
    """
    Ensure patched Agent propagates prompts/completions into the Verse generation span.
    """

    async def run_test():
        captured_kwargs = {}
        generation_ctx = DummyGenerationContext()
        sdk = Mock()
        sdk.create_generation.return_value = generation_ctx

        async def fake_events():
            if False:  # pragma: no cover - required for async generator syntax
                yield None

        def make_message(parts, finish_reason=""):
            return types.SimpleNamespace(parts=parts, finish_reason=finish_reason)

        ctx_messages = [
            make_message([make_part("SystemPromptPart", content="You are helpful")]),
            make_message([make_part("UserPromptPart", content="Hello")]),
            make_message(
                [make_part("TextPart", content="Hi there!")],
                finish_reason="stop",
            ),
        ]

        fake_ctx = types.SimpleNamespace(
            messages=ctx_messages,
            model=types.SimpleNamespace(model_name="test-model"),
            usage=types.SimpleNamespace(
                input_tokens=10,
                output_tokens=5,
                tool_calls=0,
            ),
        )

        class FakeAgent:
            _verse_patched = False

            async def run(self, *args, event_stream_handler=None, **kwargs):
                captured_kwargs["handler"] = event_stream_handler
                captured_kwargs["args"] = args
                captured_kwargs["kwargs"] = kwargs
                if event_stream_handler:
                    await event_stream_handler(fake_ctx, fake_events())
                return "ok"

            def run_stream(self, *args, **kwargs):
                return Mock()

        monkeypatch.setattr("pydantic_ai.Agent", FakeAgent)

        patch_pydantic_ai(sdk)

        agent = FakeAgent()
        result = await agent.run("ignored")

        assert result == "ok"
        assert sdk.create_generation.call_count == 1

        # prompts captured from system + user messages
        assert generation_ctx.captured_messages is not None
        prompt_roles = [m["role"] for m in generation_ctx.captured_messages]
        assert prompt_roles == ["system", "user"]

        # completions captured from assistant text
        assert generation_ctx.captured_completions is not None
        assert generation_ctx.captured_completions[0]["message"]["role"] == "assistant"
        assert (
            generation_ctx.captured_completions[0]["message"]["content"] == "Hi there!"
        )

        # finish reason propagates
        assert generation_ctx.captured_reason == "stop"
        assert generation_ctx.span.ended is True

    asyncio.run(run_test())



SYSTEM_PROMPT = """

You are an assistant that can answer questions about a fictional universe
consisting of characters, locations, occupations, items, and episodes.
You have access to a set of tools (functions) that allow you to look up and
cross-reference this information.

Always try to use the most relevant tool when a user asks about these datasets.
If the user asks something outside this domain, respond naturally without tools.

TOOLS:

1. get_character_by_id(character_id: int) -> Character
   - Returns a single character by ID.

2. get_all_characters() -> List[Character]
   - Returns all characters.

3. get_friends(character_id: int) -> List[Character]
   - Returns all friends of a character.

4. get_character_location(character_id: int) -> Location
   - Returns the location of a character.

5. get_character_occupation(character_id: int) -> Occupation
   - Returns the occupation of a character.

6. get_favorite_items(character_id: int) -> List[Item]
   - Returns all favorite items of a character.

7. get_episodes_by_character(character_id: int) -> List[Episode]
   - Returns all episodes featuring a character.

8. get_location_by_id(location_id: int) -> Location
   - Returns a location by ID.

9. get_all_locations() -> List[Location]
   - Returns all locations.

10. get_characters_by_location(location_id: int) -> List[Character]
    - Returns all characters at a specific location.

11. get_episodes_by_location(location_id: int) -> List[Episode]
    - Returns all episodes set at a specific location.

12. get_occupation_by_id(occupation_id: int) -> Occupation
    - Returns an occupation by ID.

13. get_all_occupations() -> List[Occupation]
    - Returns all occupations.

14. get_characters_by_occupation(occupation_id: int) -> List[Character]
    - Returns all characters with a given occupation.

15. get_item_by_id(item_id: int) -> Item
    - Returns an item by ID.

16. get_all_items() -> List[Item]
    - Returns all items.

17. get_characters_by_item(item_id: int) -> List[Character]
    - Returns all characters who have the item as a favorite.

18. get_episodes_by_item(item_id: int) -> List[Episode]
    - Returns all episodes featuring a specific item.

19. get_episode_by_id(episode_id: int) -> Episode
    - Returns an episode by ID.

20. get_all_episodes() -> List[Episode]
    - Returns all episodes.

21. get_characters_by_episode(episode_id: int) -> List[Character]
    - Returns all characters in a specific episode.

22. get_location_by_episode(episode_id: int) -> Location
    - Returns the location of a specific episode.

23. get_items_by_episode(episode_id: int) -> List[Item]
    - Returns all items featured in a specific episode.


USAGE RULES:
- When the user asks about characters, items, locations, occupations, or episodes,
  call the appropriate tool(s).
- If a user asks a high-level or natural-language question, decompose it into one
  or more tool calls.
- Return final answers in natural language, not raw JSON, unless the user explicitly
  requests raw data.

""".strip()



import asyncio
import atexit
import contextlib
from typing import Any, ClassVar

import litellm
from opentelemetry.trace import Status, StatusCode

from ...contexts import (
    GenerationContext,
    GenerationMessage,
)
from ...sdk import VerseSDK
from ...utils import apply_value
from .utils import (
    get_completition_choices,
    get_finish_reason,
    get_operation_type,
    get_stream_text,
    get_trace_span_from_metadata,
    get_usage,
    get_usage_from_response,
)


def configure_litellm(sdk: VerseSDK):
    """Configure Litellm to use the Verse SDK."""
    import litellm
    from litellm.integrations.custom_logger import CustomLogger

    class LitellmCallbackHandlers(CustomLogger):
        current_spans: ClassVar[dict[str, GenerationContext]] = {}

        async def async_log_failure_event(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """Handle failed async LLM requests"""
            self.log_failure_event(kwargs, response_obj, start_time, end_time)

        async def async_log_stream_event(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """Called for each streaming chunk."""
            self.log_stream_event(kwargs, response_obj, start_time, end_time)

        async def async_log_success_event(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """Handle successful async LLM requests"""
            self.log_success_event(kwargs, response_obj, start_time, end_time)

        def log_failure_event(
            self, kwargs, response_obj: Any, start_time: float, end_time: float
        ):
            """Handle failed LLM requests"""
            span_ctx = self._get_current_span(kwargs)

            if span_ctx:
                span_ctx.error(kwargs.get("exception", response_obj))
                self._drop_current_span(kwargs)

        def log_post_api_call(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """End the generation span and remove from the active context variable"""
            span_ctx = self._get_current_span(kwargs)

            if span_ctx:
                sdk.event(
                    "llm.response_received",
                    metadata={"length": len(str(response_obj))},
                    level="info",
                    span=span_ctx.span,
                )

        def log_pre_api_call(
            self, model: str, messages: list[GenerationMessage], kwargs: dict
        ):
            """Create a generation span around the LLM request"""
            id = kwargs.get("litellm_call_id")
            span = get_trace_span_from_metadata(kwargs)
            operation_type = get_operation_type(messages, kwargs)

            provider = kwargs.get("custom_llm_provider", "openai")

            generation_ctx = sdk.create_generation(
                name=f"LiteLLM: {operation_type.capitalize()}",
                messages=messages,
                metadata=kwargs.get("metadata", {}),
                model=model,
                usage=get_usage(kwargs),
                span=span,
            )

            generation_ctx.provider(provider).operation(operation_type)

            self.current_spans[id] = generation_ctx

        def log_stream_event(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """Called for each streaming chunk."""
            span_ctx = self._get_current_span(kwargs)

            if span_ctx:
                try:
                    text = get_stream_text(response_obj)
                    if text:
                        span_ctx.event(
                            "gen_ai.stream.delta",
                            metadata={"content": text[:512]},
                        )
                except Exception:
                    pass

        def log_success_event(
            self, kwargs, response_obj: Any, start_time: float, end_time: float
        ):
            """Handle successful LLM requests"""
            span_ctx = self._get_current_span(kwargs)

            if span_ctx:
                try:
                    apply_value(span_ctx, "model_used", response_obj.model)
                    apply_value(span_ctx, "reason", get_finish_reason(response_obj))

                    span_ctx.completions(
                        get_completition_choices(
                            kwargs.get("complete_streaming_response", response_obj)
                        )
                    )

                    span_ctx.usage(get_usage_from_response(response_obj))
                    span_ctx.status(Status(StatusCode.OK))
                except Exception as e:
                    span_ctx.error(e)
                finally:
                    self._drop_current_span(kwargs)

        def _drop_current_span(self, kwargs: dict):
            span = self.current_spans.pop(kwargs.get("litellm_call_id"), None)
            if span:
                span.span.end()

        def _get_current_span(self, kwargs: dict) -> GenerationContext:
            return self.current_spans.get(kwargs.get("litellm_call_id"))

    litellm.callbacks = [LitellmCallbackHandlers()]


def cleanup_litellm():
    """Clean up litellm HTTP clients to prevent garbage collection issues."""
    try:
        _cleanup_async_clients()
        _cleanup_sync_clients()
        _unregister_atexit_handlers()
    except Exception:
        pass


def _cleanup_async_clients():
    """Clean up async litellm clients."""
    if not hasattr(litellm, "close_litellm_async_clients"):
        return

    try:
        loop = _get_or_create_event_loop()
        loop.run_until_complete(litellm.close_litellm_async_clients())
    except Exception:
        pass


def _cleanup_sync_clients():
    """Clean up synchronous litellm clients."""
    if hasattr(litellm, "module_level_client") and litellm.module_level_client:
        with contextlib.suppress(Exception):
            litellm.module_level_client.close()

    if hasattr(litellm, "client_session") and litellm.client_session:
        with contextlib.suppress(Exception):
            litellm.client_session.close()


def _unregister_atexit_handlers():
    """
    Clear all atexit handlers since we've already done cleanup.
    This prevents litellm's atexit handlers from running during Python shutdown
    when logging streams are already closed, which would cause errors.
    """
    with contextlib.suppress(Exception):
        atexit._clear()


def _get_or_create_event_loop():
    """Get existing event loop or create a new one if needed."""
    try:
        loop = asyncio.get_running_loop()
        return loop
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop



from verse_sdk.exporters.verse.utils import as_traces_endpoint


class TestAsTracesEndpoint:
    def test_as_traces_endpoint_with_formatted_host(self):
        host = "https://otel.verse.com/v1/traces"
        result = as_traces_endpoint(host)
        assert result == "https://otel.verse.com/v1/traces"

    def test_as_traces_endpoint_with_host_and_api_key(self):
        host = "https://otel.verse.com"
        api_key = "1234567890"
        result = as_traces_endpoint(host, api_key)
        assert result == "https://otel.verse.com/v1/traces?api_key=1234567890"

    def test_as_traces_endpoint_with_host_and_no_api_key(self):
        host = "https://otel.verse.com"
        result = as_traces_endpoint(host, None)
        assert result == "https://otel.verse.com/v1/traces"



from __future__ import annotations

import contextlib
import gc
import logging
from collections.abc import Callable
from typing import Any

import httpcore
import httpx
from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.trace import ProxyTracerProvider

from .exporters import (
    Exporter,
    ExportersRegistry,
)
from .integrations import VENDOR_ALLOWLIST_CONFIG, apply_integrations
from .sdk import VerseSDK


class VerseSDKBuilder:
    exporters = ExportersRegistry()

    _initialized: bool = False
    _integration_cleanup: Callable | None = None
    _sdk: VerseSDK | None
    _tracer: trace.Tracer | None

    def __getattr__(self, name: str) -> Any:
        """Dynamically access methods from the SDK."""
        if not self._sdk:
            raise ValueError("SDK not initialized")

        return getattr(self._sdk, name)

    def init(
        self,
        app_name: str = "verse_sdk_observability",
        environment: str | None = "development",
        exporters: list[Exporter] | None = None,
        vendors: VENDOR_ALLOWLIST_CONFIG = None,
        version: str | None = "1.0.0",
    ) -> VerseSDK:
        """
        Lazily initialize the observability client.

        Args:
            app_name: Service name for tracing
            exporters: Exporter name(s) or None for auto-detection
            environment: Deployment environment (e.g., "production", "staging")
            version: Service version

        Returns:
            VerseSDK w/ decorations
        """
        try:
            if self._initialized:
                return self._sdk

            resource = Resource.create(
                {
                    "deployment.environment": environment,
                    "service.name": app_name,
                    "service.version": version,
                }
            )

            provider = TracerProvider(resource=resource)

            for exporter in exporters:
                try:
                    provider.add_span_processor(
                        exporter.create_span_processor(resource)
                    )
                    logging.info(
                        "Added span processor `%s`",
                        exporter.__class__.__name__,
                    )
                except Exception as e:
                    logging.warning(
                        "Error adding span processor `%s`",
                        exporter.__class__.__name__,
                        exc_info=e,
                    )

            trace.set_tracer_provider(provider)

            self._tracer = trace.get_tracer(app_name)
            self._sdk = VerseSDK(self._tracer)
            self._initialized = True
            self._integration_cleanup = apply_integrations(self._sdk, vendors)
            logging.info("🚀 VerseSDK initialized")

            return self._sdk
        except Exception as e:
            logging.error("Error initializing SDK", exc_info=e)
            raise e

    def flush(self, timeout_ms: int = 30000) -> None:
        """
        Flush all pending traces to exporters.

        Args:
            timeout_ms: Maximum time to wait for flush in milliseconds
        """
        try:
            if not self._initialized:
                return

            provider = trace.get_tracer_provider()
            if isinstance(provider, ProxyTracerProvider):
                provider.force_flush(timeout_millis=timeout_ms)
        except Exception as e:
            logging.warning("Error flushing traces", exc_info=e)

    def shutdown(self, timeout_ms: int = 30000) -> None:
        """
        Shutdown the SDK and stop all background threads.
        This should be called during application teardown to ensure clean shutdown.

        Args:
            timeout_ms: Maximum time to wait for shutdown in milliseconds
        """
        if not self._initialized:
            return

        try:
            if self._integration_cleanup:
                self._integration_cleanup()
                logging.info("Integration cleanup complete")
            else:
                logging.warning("No integration cleanup to perform")

            self.flush(timeout_ms)
            self._cleanup_http_resources()
            self._shutdown_tracer_provider(timeout_ms)
            self._reset_sdk_state()
            logging.info("VerseSDK shutdown complete")
        except Exception as e:
            logging.warning("Error shutting down SDK", exc_info=e)

    def _cleanup_http_resources(self) -> None:
        """Clean up HTTP resources to prevent memory leaks."""
        provider = trace.get_tracer_provider()
        if not isinstance(provider, TracerProvider):
            return

        try:
            self._close_exporter_clients(provider)
            self._force_close_http_objects()
        except Exception as e:
            logging.debug("Error during HTTP cleanup", exc_info=e)

    def _close_exporter_clients(self, provider: TracerProvider) -> None:
        """Close HTTP clients in span processors."""
        for processor in provider._active_span_processor._span_processors:
            try:
                if not hasattr(processor, "span_exporter"):
                    continue

                exporter = processor.span_exporter
                if hasattr(exporter, "_exporter"):
                    exporter = exporter._exporter

                if hasattr(exporter, "_session") and exporter._session:
                    exporter._session.close()

                if hasattr(exporter, "httpx_client") and exporter.httpx_client:
                    with contextlib.suppress(Exception):
                        if not exporter.httpx_client.is_closed:
                            exporter.httpx_client.close()
            except Exception as e:
                logging.debug("Error closing exporter clients", exc_info=e)

    def _force_close_http_objects(self) -> None:
        """Force close HTTP objects to prevent memory leaks during shutdown."""
        try:
            self._close_httpx_responses()
            gc.collect()
            self._close_httpcore_objects()
            self._close_httpx_clients()
        except Exception:
            pass

    def _close_httpx_responses(self) -> None:
        """Close all active httpx.Response objects."""
        responses_to_close = [
            object for object in gc.get_objects() if isinstance(object, httpx.Response)
        ]

        for response in responses_to_close:
            try:
                if not response.is_closed:
                    response.close()
            except Exception:
                pass

    def _close_httpcore_objects(self) -> None:
        """Close httpcore HTTP11 connections and connection pools."""
        for object in gc.get_objects():
            try:
                if (
                    hasattr(httpcore, "HTTP11Connection")
                    and isinstance(object, httpcore.HTTP11Connection)
                    and hasattr(object, "close")
                ) or isinstance(object, httpcore.ConnectionPool):
                    object.close()
            except Exception:
                pass

    def _close_httpx_clients(self) -> None:
        """Close all remaining httpx.Client instances."""
        for object in gc.get_objects():
            try:
                if isinstance(object, httpx.Client) and not object.is_closed:
                    if hasattr(object, "_transport") and object._transport:
                        with contextlib.suppress(Exception):
                            object._transport.close()

                    object.close()
                    logging.debug("Closed httpx.Client during shutdown")
            except Exception:
                pass

    def _shutdown_tracer_provider(self, timeout_ms: int) -> None:
        """Shutdown the tracer provider."""
        provider = trace.get_tracer_provider()
        if hasattr(provider, "shutdown") and callable(provider.shutdown):
            provider.shutdown()

    def _reset_sdk_state(self) -> None:
        """Reset SDK internal state."""
        self._initialized = False
        self._sdk = None
        self._tracer = None
        self._vendor = None



from __future__ import annotations

from typing import Any, Literal, TypedDict

FINISH_REASON_MAP = {
    "tool_use": "tool_calls",
    "end_turn": "stop",
    "max_tokens": "length",
    "stop_sequence": "stop",
}

VALID_FINISH_REASONS = {
    "content_filter",
    "error",
    "length",
    "other",
    "stop",
    "tool_calls",
}


ContextType = Literal["generation", "span", "trace"]
Level = Literal["debug", "error", "fatal", "info", "warning"]
ObservationType = Literal["observation", "trace"]
OperationType = Literal[
    "agent",
    "chain",
    "embedding",
    "evaluator",
    "event",
    "guardrail",
    "retriever",
    "tool",
]


class CompletionChoiceMessage(TypedDict):
    role: str
    content: Any
    function_call: dict | None
    name: str | None
    tool_calls: list[dict] | None


class CompletionChoice(TypedDict):
    message: CompletionChoiceMessage | None
    delta: str | None
    text: str | None


class ContextMetadata(TypedDict):
    pass


class EventMetadata(TypedDict):
    pass


class GenerationMessage(TypedDict):
    role: str
    content: str
    function_call: dict | None
    name: str | None
    tool_calls: list[dict] | None


class GenerationUsage(TypedDict):
    completion_tokens: int | None
    input_tokens: int | None
    max_tokens: int | None
    output_tokens: int | None
    prompt_tokens: int | None
    stream: bool | None
    temperature: float | None
    tool_calls: int | None
    total_tokens: int | None
    top_p: float | None


class Score(TypedDict):
    name: str
    value: float | str
    comment: str | None



from unittest.mock import Mock

import pytest

from verse_sdk.utils import (
    apply_value,
    create_basic_auth_token,
    merge,
    to_json,
)


class TestApplyValue:
    def test_apply_value_with_valid_value(self):
        mock_obj = Mock()
        apply_value(mock_obj, "set_name", "test_value")
        mock_obj.set_name.assert_called_once_with("test_value")

    def test_apply_value_with_none_value(self):
        mock_obj = Mock()
        apply_value(mock_obj, "set_name", None)
        mock_obj.set_name.assert_not_called()

    def test_apply_value_with_empty_string(self):
        mock_obj = Mock()
        apply_value(mock_obj, "set_name", "")
        mock_obj.set_name.assert_called_once_with("")

    def test_apply_value_with_zero(self):
        mock_obj = Mock()
        apply_value(mock_obj, "set_name", 0)
        mock_obj.set_name.assert_called_once_with(0)

    def test_apply_value_with_false(self):
        mock_obj = Mock()
        apply_value(mock_obj, "set_name", False)
        mock_obj.set_name.assert_called_once_with(False)

    def test_apply_value_with_list(self):
        mock_obj = Mock()
        test_list = [1, 2, 3]
        apply_value(mock_obj, "set_items", test_list)
        mock_obj.set_items.assert_called_once_with(test_list)

    def test_apply_value_with_dict(self):
        mock_obj = Mock()
        test_dict = {"key": "value"}
        apply_value(mock_obj, "set_config", test_dict)
        mock_obj.set_config.assert_called_once_with(test_dict)


class TestCreateBasicAuthToken:
    def test_create_basic_auth_token(self):
        result = create_basic_auth_token("user", "pass")
        assert result == "Basic dXNlcjpwYXNz"

    def test_create_basic_auth_token_with_special_chars(self):
        result = create_basic_auth_token("user@domain", "pass:word")
        assert result == "Basic dXNlckBkb21haW46cGFzczp3b3Jk"

    def test_create_basic_auth_token_with_empty_strings(self):
        result = create_basic_auth_token("", "")
        assert result == "Basic Og=="

    def test_create_basic_auth_token_with_unicode(self):
        result = create_basic_auth_token("测试", "密码")
        assert result == "Basic 5rWL6K+VOuWvhueggQ=="

    def test_create_basic_auth_token_with_numbers(self):
        result = create_basic_auth_token("123", "456")
        assert result == "Basic MTIzOjQ1Ng=="

    def test_create_basic_auth_token_with_spaces(self):
        result = create_basic_auth_token("user name", "pass word")
        assert result == "Basic dXNlciBuYW1lOnBhc3Mgd29yZA=="


class TestMerge:
    def test_merge_both_none(self):
        result = merge(None, None)
        assert result == {}

    def test_merge_a_none_b_empty(self):
        result = merge(None, {})
        assert result == {}

    def test_merge_a_none_b_with_data(self):
        b = {"key1": "value1", "key2": "value2"}
        result = merge(None, b)
        assert result == {"key1": "value1", "key2": "value2"}
        assert result is not b

    def test_merge_a_empty_b_none(self):
        result = merge({}, None)
        assert result == {}

    def test_merge_a_with_data_b_none(self):
        a = {"key1": "value1", "key2": "value2"}
        result = merge(a, None)
        assert result == {"key1": "value1", "key2": "value2"}
        assert result is not a

    def test_merge_both_with_data(self):
        a = {"key1": "value1", "key2": "value2"}
        b = {"key2": "new_value2", "key3": "value3"}
        result = merge(a, b)
        assert result == {"key1": "value1", "key2": "new_value2", "key3": "value3"}

    def test_merge_b_overrides_a(self):
        a = {"key1": "value1", "key2": "value2"}
        b = {"key2": "override", "key3": "value3"}
        result = merge(a, b)
        assert result["key2"] == "override"

    def test_merge_with_nested_dicts(self):
        a = {"config": {"host": "localhost", "port": 8080}}
        b = {"config": {"port": 9090, "ssl": True}}
        result = merge(a, b)
        assert result == {"config": {"port": 9090, "ssl": True}}

    def test_merge_with_different_types(self):
        a = {"string": "value", "number": 42, "boolean": True}
        b = {"string": "new_value", "list": [1, 2, 3]}
        result = merge(a, b)
        assert result == {
            "string": "new_value",
            "number": 42,
            "boolean": True,
            "list": [1, 2, 3],
        }

    def test_merge_preserves_original_dicts(self):
        a = {"key1": "value1"}
        b = {"key2": "value2"}
        original_a = a.copy()
        original_b = b.copy()

        result = merge(a, b)

        assert a == original_a
        assert b == original_b
        assert result == {"key1": "value1", "key2": "value2"}


class TestToJson:
    def test_to_json_with_simple_dict(self):
        data = {"key": "value", "number": 42}
        result = to_json(data)
        assert result == '{"key": "value", "number": 42}'

    def test_to_json_with_nested_dict(self):
        data = {"config": {"host": "localhost", "port": 8080}}
        result = to_json(data)
        assert result == '{"config": {"host": "localhost", "port": 8080}}'

    def test_to_json_with_list(self):
        data = {"items": [1, 2, 3], "name": "test"}
        result = to_json(data)
        assert result == '{"items": [1, 2, 3], "name": "test"}'

    def test_to_json_with_boolean(self):
        data = {"enabled": True, "disabled": False}
        result = to_json(data)
        assert result == '{"enabled": true, "disabled": false}'

    def test_to_json_with_none_values(self):
        data = {"key1": "value", "key2": None}
        result = to_json(data)
        assert result == '{"key1": "value", "key2": null}'

    def test_to_json_with_empty_dict(self):
        data = {}
        result = to_json(data)
        assert result == "{}"

    def test_to_json_with_unicode(self):
        data = {"message": "Hello 世界", "emoji": "🚀"}
        result = to_json(data)
        assert (
            result == '{"message": "Hello \\u4e16\\u754c", "emoji": "\\ud83d\\ude80"}'
        )

    def test_to_json_with_non_serializable_object(self):
        class NonSerializable:
            def __repr__(self):
                return "NonSerializable()"

        data = {"obj": NonSerializable()}
        result = to_json(data)
        assert result == "{'obj': NonSerializable()}"

    def test_to_json_with_circular_reference(self):
        data = {"key": "value"}
        data["self"] = data
        result = to_json(data)
        assert "key" in result
        assert "value" in result
        assert "..." in result

    def test_to_json_with_custom_object(self):
        class CustomObject:
            def __init__(self, value):
                self.value = value

            def __repr__(self):
                return f"CustomObject({self.value})"

        data = {"custom": CustomObject("test")}
        result = to_json(data)
        assert result == "{'custom': CustomObject(test)}"

    def test_to_json_with_exception_during_serialization(self):
        class FailingObject:
            def __repr__(self):
                raise Exception("Repr failed")

        data = {"failing": FailingObject()}
        with pytest.raises(Exception, match="Repr failed"):
            to_json(data)



import inspect
import os
from collections.abc import Callable
from typing import Any

from anthropic import Anthropic

from tests.fixtures.agents.prompt import SYSTEM_PROMPT
from tests.fixtures.datasource import datasource


def convert_functions_to_tools(functions: list[Callable]):
    tools = []
    for func in functions:
        sig = inspect.signature(func)

        properties = {}
        required = []

        for param_name, param in sig.parameters.items():
            if param_name in ("self", "cls"):
                continue

            param_type = param.annotation
            json_type = _python_type_to_json_type(param_type)

            properties[param_name] = {"type": json_type}

            if param.default == inspect.Parameter.empty:
                required.append(param_name)

        tools.append(
            {
                "name": func.__name__,
                "description": func.__doc__ or f"Function {func.__name__}",
                "input_schema": {
                    "type": "object",
                    "properties": properties,
                    "required": required,
                },
            }
        )

    return tools


def _python_type_to_json_type(python_type):
    type_mapping = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
    }

    origin = getattr(python_type, "__origin__", None)
    if origin is not None:
        return type_mapping.get(origin, "string")

    return type_mapping.get(python_type, "string")


class AnthropicAgent:
    def __init__(self):
        self.client = Anthropic(
            api_key=os.getenv("ANTHROPIC_API_KEY"),
        )

    def ask(self, question: str) -> list[Any]:
        response = self.client.messages.create(
            max_tokens=500,
            model="claude-sonnet-4-0",
            messages=[
                {
                    "role": "user",
                    "content": question,
                },
            ],
            tools=convert_functions_to_tools(datasource),
            system=SYSTEM_PROMPT,
        )

        return response.content

    async def stream(self, question: str):
        """
        Stream responses from Anthropic with tool call support.

        Args:
            question: The question to ask

        Yields:
            Text chunks from the streaming response
        """
        messages = [
            {
                "role": "user",
                "content": question,
            },
        ]

        max_iterations = 3
        for _ in range(max_iterations):
            with self.client.messages.stream(
                max_tokens=500,
                model="claude-sonnet-4-0",
                messages=messages,
                tools=convert_functions_to_tools(datasource),
                system=SYSTEM_PROMPT,
            ) as stream:
                tool_use_blocks = []
                accumulated_text = []

                for event in stream:
                    if (
                        event.type == "content_block_delta"
                        and hasattr(event.delta, "text")
                        and event.delta.text is not None
                    ):
                        accumulated_text.append(event.delta.text)
                        yield event.delta.text
                    elif (
                        event.type == "content_block_start"
                        and hasattr(event.content_block, "type")
                        and event.content_block.type == "tool_use"
                    ):
                        tool_use_blocks.append(event.content_block)
                    elif (
                        event.type == "content_block_delta"
                        and hasattr(event.delta, "type")
                        and event.delta.type == "tool_use"
                        and hasattr(event.delta, "partial_json")
                    ):
                        pass

                final_message = stream.get_final_message()
                if final_message and hasattr(final_message, "content"):
                    tool_results = []

                    for block in final_message.content:
                        if hasattr(block, "type") and block.type == "tool_use":
                            tool_name = block.name
                            tool_input = block.input

                            func = None
                            for f in datasource:
                                if f.__name__ == tool_name:
                                    func = f
                                    break

                            if func:
                                try:
                                    if isinstance(tool_input, dict):
                                        result = func(**tool_input)
                                    else:
                                        result = func(tool_input)
                                    tool_results.append(
                                        {
                                            "type": "tool_result",
                                            "tool_use_id": block.id,
                                            "content": str(result),
                                        }
                                    )
                                except Exception as e:
                                    error_msg = repr(e)
                                    tool_results.append(
                                        {
                                            "type": "tool_result",
                                            "tool_use_id": block.id,
                                            "content": f"Error: {error_msg}",
                                        }
                                    )

                    if tool_results:
                        messages.append(
                            {
                                "role": "assistant",
                                "content": final_message.content,
                            }
                        )
                        messages.append({"role": "user", "content": tool_results})
                        continue

                break



import json
import logging
from typing import Any

from pydantic_ai import RunContext, RunUsage

from ...contexts import CompletionChoice, GenerationMessage, GenerationUsage
from ...utils import get
from .constants import VENDOR_ANTHROPIC, VENDOR_OPENAI, VENDOR_UNKNOWN


def calculate_total_tokens(usage: RunUsage) -> int | None:
    """Calculate the total tokens."""
    return (
        usage.input_tokens + usage.output_tokens
        if usage.input_tokens and usage.output_tokens
        else None
    )


def get_finish_reason(ctx: RunContext) -> str:
    """Get the finish reason from RunContext."""
    if hasattr(ctx, "messages") and ctx.messages:
        return get(ctx.messages[-1], "finish_reason", "")

    return ""


def get_generation_messages(
    ctx: RunContext,
) -> tuple[list[GenerationMessage], list[CompletionChoice]]:
    """
    Convert RunContext messages into prompt/completion payloads.

    Prompts capture system/user/tool responses. Completions capture assistant output.
    """
    prompts: list[GenerationMessage] = []
    completions: list[CompletionChoice] = []

    if not hasattr(ctx, "messages") or not ctx.messages:
        return prompts, completions

    for message in ctx.messages:
        if not hasattr(message, "parts"):
            continue

        assistant_content_parts = []
        assistant_tool_calls = []

        for part in message.parts:
            part_type = part.__class__.__name__

            if part_type == "SystemPromptPart":
                prompts.append(
                    GenerationMessage(
                        content=part.content,
                        function_call=None,
                        name=None,
                        role="system",
                        tool_calls=None,
                    )
                )
            elif part_type == "UserPromptPart":
                prompts.append(
                    GenerationMessage(
                        content=part.content,
                        function_call=None,
                        name=None,
                        role="user",
                        tool_calls=None,
                    )
                )
            elif part_type == "ToolReturnPart":
                prompts.append(
                    GenerationMessage(
                        content=json.dumps(part.content),
                        function_call=None,
                        name=None,
                        role="tool",
                        tool_calls=None,
                        id=get(part, "tool_call_id"),
                    )
                )
            elif part_type == "TextPart":
                assistant_content_parts.append(part.content)
            elif part_type == "ToolCallPart":
                assistant_tool_calls.append(
                    {
                        "function": {
                            "arguments": part.args,
                            "name": part.tool_name,
                        },
                        "id": part.tool_call_id,
                        "type": "function",
                    }
                )

        if assistant_tool_calls or assistant_content_parts:
            message_content = (
                " ".join(assistant_content_parts) if assistant_content_parts else ""
            )
            completions.append(
                CompletionChoice(
                    message={
                        "content": message_content,
                        "function_call": None,
                        "name": None,
                        "role": "assistant",
                        "tool_calls": assistant_tool_calls or None,
                    },
                    delta=None,
                    text=None,
                )
            )

    return prompts, completions


def get_llm_info_from_model(model: Any) -> tuple[str, str]:
    """Get the vendor and model name."""
    try:
        if isinstance(model, get_model_instance(VENDOR_OPENAI)):
            return VENDOR_OPENAI, model.model_name
        if isinstance(model, get_model_instance(VENDOR_ANTHROPIC)):
            return VENDOR_ANTHROPIC, model.model_name

        return VENDOR_UNKNOWN, getattr(model, "model_name", VENDOR_UNKNOWN)
    except Exception:
        return VENDOR_UNKNOWN, VENDOR_UNKNOWN


def get_model_instance(model_name: str) -> Any:
    """Lazy load the vendor model by name."""
    try:
        if model_name.startswith(VENDOR_OPENAI):
            from pydantic_ai.models.openai import OpenAIChatModel

            return OpenAIChatModel
        if model_name.startswith(VENDOR_ANTHROPIC):
            from pydantic_ai.models.anthropic import AnthropicChatModel

            return AnthropicChatModel

        return None
    except Exception:
        logging.warning(f"Error loading model {model_name}", exc_info=True)
        return None


def get_operation_type(ctx: RunContext) -> str:
    """Get the operation type based on the messages."""
    if hasattr(ctx, "messages") and ctx.messages:
        return "chat"
    return "completion"


def get_usage(usage: RunUsage) -> GenerationUsage:
    """Get the usage from RunContext."""
    return GenerationUsage(
        completion_tokens=usage.output_tokens,
        input_tokens=usage.input_tokens,
        output_tokens=usage.output_tokens,
        prompt_tokens=usage.input_tokens,
        tool_calls=usage.tool_calls,
        total_tokens=calculate_total_tokens(usage),
    )



import functools
import logging
from collections.abc import AsyncIterable

from ...sdk import VerseSDK


def patch_pydantic_ai(sdk: VerseSDK):
    """Patch Pydantic AI to use the Verse SDK."""

    from pydantic_ai import Agent, RunContext
    from pydantic_ai.messages import (
        AgentStreamEvent,
        FunctionToolResultEvent,
        PartDeltaEvent,
    )

    from .utils import (
        get_finish_reason,
        get_generation_messages,
        get_llm_info_from_model,
        get_operation_type,
        get_usage,
    )

    async def handle_events(ctx: RunContext, events: AsyncIterable[AgentStreamEvent]):
        """
        Handle Pydantic events to create LLM generation spans.
        Pydantic dispatches more events than we need, so the delta and finish reason checks are necessary to avoid duplicates.
        """
        try:
            contains_generation = get_finish_reason(ctx) == "stop"

            async for event in events:
                if isinstance(event, (PartDeltaEvent, FunctionToolResultEvent)):
                    contains_generation = True

            if not contains_generation:
                return

            op = get_operation_type(ctx)
            vendor, model_name = get_llm_info_from_model(ctx.model)
            prompts, completions = get_generation_messages(ctx)

            generation_ctx = sdk.create_generation(
                f"Pydantic AI: {op.capitalize()}",
                model=model_name,
                usage=get_usage(ctx.usage),
                vendor=vendor,
            )

            generation_ctx.provider(vendor).operation(op)

            if prompts:
                generation_ctx.messages(prompts)

            if completions:
                generation_ctx.completions(completions)

            finish_reason = get_finish_reason(ctx)
            if finish_reason:
                generation_ctx.reason(finish_reason)

            generation_ctx.span.end()
        except Exception as e:
            logging.warning("Pydantic streaming event dropped", exc_info=e)

    def patch():
        """Setup monkey-patches for Pydantic AI if not already patched."""
        if not getattr(Agent, "_verse_patched", False):
            patch_run()
            patch_run_stream()
            Agent._verse_patched = True

    def patch_run():
        """Setup a monkey patch for Agent.run() that captures the LLM call in a generation span."""
        original_run_method = Agent.run

        @functools.wraps(original_run_method)
        async def run__patched(self, *args, **kwargs):
            try:
                return await original_run_method(
                    self, *args, **kwargs, event_stream_handler=handle_events
                )
            except Exception as e:
                logging.warning("Error patching Agent.run()", exc_info=e)
                return original_run_method(self, *args, **kwargs)

        Agent.run = run__patched

    def patch_run_stream():
        """Setup a monkey patch for Agent.run_stream() that captures streaming LLM calls."""
        original_run_stream_method = Agent.run_stream

        @functools.wraps(original_run_stream_method)
        def run_stream__patched(self, *args, **kwargs):
            try:
                stream = original_run_stream_method(self, *args, **kwargs)
                return _wrap_stream_context_manager(stream, self, sdk)
            except Exception as e:
                logging.warning("Error patching Agent.run_stream()", exc_info=e)
                return original_run_stream_method(self, *args, **kwargs)

        Agent.run_stream = run_stream__patched

    class TracedStreamContextManager:
        """Wrapper for run_stream() context manager that traces streaming events."""

        def __init__(self, stream, agent, sdk: VerseSDK):
            self._stream = stream
            self._agent = agent
            self._sdk = sdk
            self._inner_stream = None
            self._entered = False

        async def __aenter__(self):
            self._inner_stream = await self._stream.__aenter__()
            self._entered = True
            return self

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            return await self._stream.__aexit__(exc_type, exc_val, exc_tb)

        def __aiter__(self):
            if not self._entered:
                raise RuntimeError(
                    "TracedStreamContextManager must be entered with 'async with' before iteration"
                )
            return _wrap_stream_iterator(self._inner_stream, self._agent, self._sdk)

        def stream_output(self):
            """Wrap stream_output() to trace events."""
            if not self._entered:
                raise RuntimeError(
                    "TracedStreamContextManager must be entered with 'async with' before calling stream_output()"
                )
            original_stream = self._inner_stream.stream_output()
            return _wrap_stream_iterator(original_stream, self._agent, self._sdk)

        def __getattr__(self, name):
            """Delegate attribute access to the underlying stream result."""
            if self._inner_stream is None:
                raise AttributeError(
                    f"'{type(self).__name__}' object has no attribute '{name}'. "
                    "Context manager must be entered first."
                )
            return getattr(self._inner_stream, name)

    def _wrap_stream_context_manager(stream, agent, sdk: VerseSDK):
        """Wrap the stream context manager to return a traced version."""
        return TracedStreamContextManager(stream, agent, sdk)

    async def _wrap_stream_iterator(stream_result, agent, sdk: VerseSDK):
        """Wrap the stream iterator to trace events as they come in."""
        generation_ctx = None
        first_event = True
        accumulated_content = []

        try:
            if hasattr(stream_result, "stream_output"):
                event_iter = stream_result.stream_output()
            elif hasattr(stream_result, "stream"):
                event_iter = stream_result.stream()
            elif hasattr(stream_result, "__aiter__"):
                event_iter = stream_result
            else:
                raise TypeError(f"Cannot iterate over {type(stream_result)}")

            async for event in event_iter:
                if first_event:
                    try:
                        op = "chat"
                        vendor, model_name = get_llm_info_from_model(agent.model)

                        generation_ctx = sdk.create_generation(
                            f"Pydantic AI: {op.capitalize()}",
                            model=model_name,
                            vendor=vendor,
                        )

                        generation_ctx.provider(vendor).operation(op)
                        first_event = False
                    except Exception as e:
                        logging.warning(
                            "Error initializing generation context for stream",
                            exc_info=e,
                        )

                if isinstance(event, PartDeltaEvent) and generation_ctx:
                    try:
                        if hasattr(event, "delta"):
                            delta = event.delta
                            if hasattr(delta, "content") and delta.content:
                                content_str = str(delta.content)
                                accumulated_content.append(content_str)
                                sdk.event(
                                    "gen_ai.stream.delta",
                                    metadata={"content": content_str[:512]},
                                    level="info",
                                    span=generation_ctx.span,
                                )
                    except Exception as e:
                        logging.warning("Error processing delta event", exc_info=e)

                yield event

            if generation_ctx:
                try:
                    if accumulated_content:
                        from ...contexts import CompletionChoice

                        full_content = "".join(accumulated_content)
                        generation_ctx.completions(
                            [
                                CompletionChoice(
                                    message={
                                        "role": "assistant",
                                        "content": full_content,
                                        "function_call": None,
                                        "name": None,
                                        "tool_calls": None,
                                    },
                                    delta=None,
                                    text=full_content,
                                )
                            ]
                        )
                except Exception as e:
                    logging.warning("Error finalizing stream span", exc_info=e)
                finally:
                    generation_ctx.span.end()

        except Exception as e:
            if generation_ctx:
                generation_ctx.error(e)
                generation_ctx.span.end()
            raise

    patch()



from __future__ import annotations

import base64
import json
from typing import Any


def apply_value(obj: Any, method_name: str, value: Any) -> None:
    """Apply a value to an object using a method name."""
    if value is not None:
        getattr(obj, method_name)(value)


def create_basic_auth_token(key: str, secret: str) -> str:
    """Create a basic authentication token."""
    token = base64.b64encode(f"{key}:{secret}".encode()).decode()
    return f"Basic {token}"


def format_hex_id(hex_id: str) -> int:
    """Format a span ID."""
    trace_id_hex = hex_id.replace("0x", "").replace("0X", "")
    return int(trace_id_hex, 16)


def get(target: Any, key: str, default_value: Any = "") -> Any:
    """Get a value from an object using a key."""
    if target is None:
        return default_value
    if isinstance(target, dict):
        return target.get(key, default_value)
    elif hasattr(target, key):
        return getattr(target, key) or default_value
    return default_value


def merge(a: dict[str, Any] | None, b: dict[str, Any] | None) -> dict[str, Any]:
    """Merge two dictionaries, handling None values."""
    if a is None and b is None:
        return {}
    if a is None:
        return b.copy() if b else {}
    if b is None:
        return a.copy() if a else {}
    return {**a, **b}


def to_json(data: dict[str, Any]) -> str:
    """Convert a dictionary to a JSON string."""
    try:
        return json.dumps(data)
    except Exception:
        return repr(data)


def uses_pydantic_base_model(model: Any) -> bool:
    """Check if the model is a Pydantic BaseModel."""
    try:
        from pydantic import BaseModel

        return isinstance(model, BaseModel)
    except ImportError:
        return False



from __future__ import annotations

from unittest.mock import patch

from verse_sdk.exporters import ExportersRegistry


class TestExportersRegistry:
    def setup_method(self):
        self.registry = ExportersRegistry()

    def test_console_exporter_with_no_config(self):
        exporter = self.registry.console()

        assert exporter is not None
        assert hasattr(exporter, "config")
        assert exporter.config == {}

    def test_console_exporter_with_config(self):
        config = {"scopes": ["test_scope"]}
        exporter = self.registry.console(config)

        assert exporter is not None
        assert hasattr(exporter, "config")
        assert exporter.config == config

    def test_console_exporter_with_none_config(self):
        exporter = self.registry.console(None)

        assert exporter is not None
        assert hasattr(exporter, "config")
        assert exporter.config == {}

    def test_langfuse_exporter_with_config(self):
        config = {
            "public_key": "custom_public_key",
            "secret_key": "custom_secret_key",
            "host": "custom_host",
            "region": "us",
            "scopes": ["test_scope"],
        }

        exporter = self.registry.langfuse(config)

        assert exporter is not None
        assert hasattr(exporter, "config")
        assert exporter.config["public_key"] == "custom_public_key"
        assert exporter.config["secret_key"] == "custom_secret_key"

    def test_langfuse_exporter_validation_error(self):
        config = {
            "public_key": "test_public_key",
            "secret_key": None,
        }

        exporter = self.registry.langfuse(config)

        assert exporter is None

    def test_safe_named_exporter_decorator_success(self):
        @ExportersRegistry.safe_named_exporter
        def test_function():
            return "success"

        result = test_function()
        assert result == "success"

    def test_safe_named_exporter_decorator_exception(self):
        @ExportersRegistry.safe_named_exporter
        def test_function():
            raise ValueError("Test error")

        with patch("verse_sdk.exporters.logging") as mock_logging:
            result = test_function()

            assert result is None
            mock_logging.warning.assert_called_once()

    def test_safe_named_exporter_decorator_preserves_function_metadata(self):
        @ExportersRegistry.safe_named_exporter
        def test_function():
            return "success"

        assert test_function.__name__ == "test_function"
        assert test_function.__doc__ is None

    def test_registry_instance_methods(self):
        console_exporter = self.registry.console()
        assert console_exporter is not None

        langfuse_exporter = self.registry.langfuse(
            {"public_key": "test", "secret_key": "test"}
        )
        assert langfuse_exporter is not None

    def test_registry_class_methods(self):
        registry = ExportersRegistry()
        console_exporter = registry.console()
        assert console_exporter is not None

        langfuse_exporter = registry.langfuse(
            {"public_key": "test", "secret_key": "test"}
        )
        assert langfuse_exporter is not None



import json

from tests.fixtures.spans import get_span_json
from verse_sdk import OtelGenAISpecCheck


class TestSpecCheckMetadata:
    def test_metadata_structure(self):
        span = get_span_json("valid_chat_complete")
        checker = OtelGenAISpecCheck(span)
        metadata = checker.extract_metadata()
        assert metadata["environment"] == "dev"
        assert metadata["model_provider"] == "anthropic"
        assert metadata["model"] == "claude-3-5-sonnet-20241022"
        assert metadata["project_id"] == "proj_1234567890"
        assert metadata["session_id"] == "sess_1234567890"
        assert metadata["start_time"] is not None
        assert metadata["end_time"] is not None
        assert metadata["span_id"] is not None
        assert metadata["span_parent_id"] is not None
        assert metadata["trace_id"] is not None
        assert metadata["user_id"] == "user_1234567890"


class TestValidSpans:
    def test_valid_chat_complete_has_all_common_fields(self):
        span = get_span_json("valid_chat_complete")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is False
        assert checker.is_attribute_valid("gen_ai.operation.name") is True

    def test_embeddings_complete_has_dimension_count(self):
        span = get_span_json("embeddings_complete")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_deprecations is False
        assert checker.is_attribute_valid("gen_ai.embeddings.dimension.count") is True

    def test_tool_execution_has_tool_fields(self):
        span = get_span_json("tool_execution")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_deprecations is False
        assert checker.is_attribute_valid("gen_ai.tool.name") is True
        assert checker.is_attribute_valid("gen_ai.tool.type") is True
        assert checker.is_attribute_valid("gen_ai.tool.description") is True


class TestMissingRequiredFields:
    def test_missing_operation_and_provider(self):
        span = get_span_json("missing_required_fields")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.operation.name") is True
        assert checker.is_attribute_invalid("gen_ai.provider.name") is True


class TestMissingRecommendedFields:
    def test_missing_recommended_fields_shows_warnings(self):
        span = get_span_json("missing_recommended_fields")
        checker = OtelGenAISpecCheck(span)

        assert checker.could_improve is True
        assert checker.is_attribute_missing("gen_ai.response.model") is True
        assert checker.is_attribute_missing("gen_ai.usage.input_tokens") is True


class TestInvalidEnumValues:
    def test_invalid_operation_name(self):
        span = get_span_json("invalid_enum_values")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.operation.name") is True

    def test_invalid_provider_name(self):
        span = get_span_json("invalid_enum_values")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.provider.name") is True

    def test_invalid_finish_reason(self):
        span = get_span_json("invalid_enum_values")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.response.finish_reasons") is True

    def test_invalid_output_type(self):
        span = get_span_json("invalid_enum_values")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.output.type") is True


class TestDeprecatedAttributes:
    def test_deprecated(self):
        span = get_span_json("deprecated_attributes")
        checker = OtelGenAISpecCheck(span)
        assert checker.is_attribute_deprecated("gen_ai.system") is True
        assert checker.is_attribute_deprecated("gen_ai.prompt") is True
        assert checker.is_attribute_deprecated("gen_ai.completion") is True


class TestInvalidInput:
    def test_invalid_json_string(self):
        with open("tests/fixtures/spans/invalid_json.txt") as f:
            invalid_json = f.read()

        checker = OtelGenAISpecCheck(invalid_json)
        assert checker.has_errors is True

    def test_dict_input(self):
        span_dict = json.loads(get_span_json("valid_chat_complete"))
        checker = OtelGenAISpecCheck(span_dict)
        assert checker.has_errors is False


class TestInvalidTypes:
    def test_invalid_string_type(self):
        span = get_span_json("invalid_types")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.operation.name") is True

    def test_invalid_int_type(self):
        span = get_span_json("invalid_types")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.request.max_tokens") is True
        assert checker.is_attribute_invalid("gen_ai.usage.input_tokens") is True
        assert checker.is_attribute_invalid("server.port") is True

    def test_invalid_double_type(self):
        span = get_span_json("invalid_types")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.request.temperature") is True

    def test_invalid_string_array_type(self):
        span = get_span_json("invalid_types")
        checker = OtelGenAISpecCheck(span)

        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.response.finish_reasons") is True

    def test_valid_attributes_are_still_valid(self):
        span = get_span_json("invalid_types")
        checker = OtelGenAISpecCheck(span)

        # These attributes have correct types and should be valid
        assert checker.is_attribute_valid("gen_ai.provider.name") is True
        assert checker.is_attribute_valid("gen_ai.request.top_p") is True
        assert checker.is_attribute_valid("gen_ai.usage.output_tokens") is True


class TestValidationDetails:
    def test_validation_details_structure(self):
        span = get_span_json("missing_required_fields")
        checker = OtelGenAISpecCheck(span)
        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.operation.name") is True
        assert checker.is_attribute_invalid("gen_ai.provider.name") is True

    def test_error_count_matches_errors(self):
        span = get_span_json("invalid_enum_values")
        checker = OtelGenAISpecCheck(span)
        assert checker.has_errors is True
        assert checker.is_attribute_invalid("gen_ai.operation.name") is True
        assert checker.is_attribute_invalid("gen_ai.provider.name") is True



from typing import Literal

from tests.fixtures.agents.anthropic_agent import AnthropicAgent
from tests.fixtures.agents.base import AgentProtocol
from tests.fixtures.agents.google_agent import GoogleAgent
from tests.fixtures.agents.langchain_agent import LangchainAgent
from tests.fixtures.agents.litellm_agent import LitellmAgent
from tests.fixtures.agents.pydantic_agent import PydanticAgent

VendorType = Literal["anthropic", "google", "langchain", "litellm", "pydantic_ai"]


class AgentFixtures:
    """
    Factory for creating standardized test agents.

    All agents implement the AgentProtocol interface with:
    - ask(question: str) -> str: Synchronous question answering
    - stream(question: str) -> AsyncIterator[str]: Async streaming responses

    This makes tests plug-and-play across different vendor integrations.
    """

    def create(self, vendor: VendorType) -> AgentProtocol:
        """
        Create an agent for the specified vendor.

        Args:
            vendor: The vendor integration to test

        Returns:
            An agent implementing the AgentProtocol interface
        """
        match vendor:
            case "anthropic":
                return AnthropicAgent()
            case "google":
                return GoogleAgent()
            case "langchain":
                return LangchainAgent()
            case "litellm":
                return LitellmAgent()
            case "pydantic_ai":
                return PydanticAgent()
            case _:
                msg = f"Unknown vendor: {vendor}"
                raise ValueError(msg)

    # Convenience methods for direct access
    def create_anthropic_agent(self) -> AnthropicAgent:
        """Create an Anthropic agent"""
        return AnthropicAgent()

    def create_litellm_agent(self) -> LitellmAgent:
        """Create a LiteLLM agent"""
        return LitellmAgent()

    def create_pydantic_agent(self) -> PydanticAgent:
        """Create a Pydantic AI agent"""
        return PydanticAgent()

    def create_langchain_agent(self) -> LangchainAgent:
        """Create a LangChain agent"""
        return LangchainAgent()

    def create_google_agent(self) -> GoogleAgent:
        """Create a Google GenAI agent"""
        return GoogleAgent()



import logging
import os

import pytest
from dotenv import load_dotenv

from verse_sdk import verse
from verse_sdk.exporters.file import FileConfig

load_dotenv()


verse.init(
    app_name="verse_e2e_suite",
    environment="development",
    exporters=[
        verse.exporters.file(FileConfig(folder_path=".files")),
        verse.exporters.console(),
        # Note:
        # Add more exporters to test remote origins (e.g. Langfuse, OTLP, etc.)
        # On an as-needed basis without committing them to the repo
    ],
    vendors=[
        "anthropic",
        "google",
        "langchain",
        "litellm",
        "pydantic_ai",
    ],
    version="1.0.0",
)


@pytest.fixture(scope="session", autouse=True)
def cleanup_files():
    yield

    try:
        for file in os.listdir(".files/"):
            os.remove(f".files/{file}")
    except Exception as e:
        logging.error(f"Error cleaning up files: {e}")



from __future__ import annotations

from unittest.mock import Mock, patch

import pytest

from verse_sdk.contexts.base import BaseContext
from verse_sdk.contexts.types import (
    ContextMetadata,
    ContextType,
    ObservationType,
    Score,
)


class ConcreteContext(BaseContext):
    @property
    def observation_type(self) -> ObservationType:
        return "observation"

    @property
    def type(self) -> ContextType:
        return "span"


class TestBaseContext:
    def setup_method(self):
        self.mock_span = Mock()
        self.mock_span.add_event = Mock()
        self.mock_span.get_span_context.return_value.is_valid = True
        self.mock_span.record_exception = Mock()
        self.mock_span.set_attribute = Mock()
        self.mock_span.set_status = Mock()

    def test_init_with_valid_span(self):
        context = ConcreteContext(self.mock_span)

        assert context._span == self.mock_span
        self.mock_span.set_attribute.assert_any_call("context.type", "span")
        self.mock_span.set_attribute.assert_any_call("observation.type", "observation")

    def test_init_with_invalid_span_raises_error(self):
        invalid_span = Mock()
        invalid_span.get_span_context.return_value.is_valid = False

        with patch("verse_sdk.contexts.base.logging"), pytest.raises(
            ValueError, match="Span invalid"
        ):
            ConcreteContext(invalid_span)

    def test_init_with_none_span_raises_error(self):
        with patch("verse_sdk.contexts.base.logging"), pytest.raises(
            ValueError, match="Span invalid"
        ):
            ConcreteContext(None)

    def test_span_property(self):
        context = ConcreteContext(self.mock_span)
        assert context.span == self.mock_span

    def test_observation_type_property_not_implemented(self):
        class IncompleteContext(BaseContext):
            @property
            def type(self) -> ContextType:
                return "span"

        with pytest.raises(
            NotImplementedError, match="Observation type must be implemented"
        ):
            assert IncompleteContext(self.mock_span).observation_type

    def test_type_property_not_implemented(self):
        class IncompleteContext(BaseContext):
            pass

        with pytest.raises(
            NotImplementedError, match="Context type must be implemented"
        ):
            assert IncompleteContext(self.mock_span).type

    def test_metadata_success(self):
        context = ConcreteContext(self.mock_span)
        metadata: ContextMetadata = {"key1": "value1", "key2": "value2"}

        result = context.metadata(metadata)
        assert result == context
        self.mock_span.set_attribute.assert_any_call("metadata.key1", "value1")
        self.mock_span.set_attribute.assert_any_call("metadata.key2", "value2")

    def test_metadata_with_none_values(self):
        context = ConcreteContext(self.mock_span)
        metadata: ContextMetadata = {"key1": "value1", "key2": None, "key3": "value3"}

        context.metadata(metadata)

        self.mock_span.set_attribute.assert_any_call("metadata.key1", "value1")
        self.mock_span.set_attribute.assert_any_call("metadata.key3", "value3")
        assert not any(
            call[0][0] == "metadata.key2"
            for call in self.mock_span.set_attribute.call_args_list
        )

    def test_error_success(self):
        context = ConcreteContext(self.mock_span)
        exception = ValueError("Test error message")

        result = context.error(exception)

        assert result == context
        self.mock_span.set_attribute.assert_any_call("error", True)
        self.mock_span.set_attribute.assert_any_call("error.type", "ValueError")
        self.mock_span.set_attribute.assert_any_call(
            "error.message", "Test error message"
        )
        self.mock_span.record_exception.assert_called_once_with(exception)
        self.mock_span.set_status.assert_called_once()

    def test_error_with_exception_during_processing(self):
        context = ConcreteContext(self.mock_span)
        exception = ValueError("Test error message")

        self.mock_span.set_attribute.side_effect = Exception("Attribute setting failed")

        with patch("verse_sdk.contexts.base.logging") as mock_logging:
            result = context.error(exception)

            assert result == context
            mock_logging.warning.assert_called_once()

    def test_score_generation_context(self):
        class GenerationContext(BaseContext):
            @property
            def observation_type(self) -> ObservationType:
                return "observation"

            @property
            def type(self) -> ContextType:
                return "generation"

        context = GenerationContext(self.mock_span)
        score: Score = {
            "name": "quality",
            "value": 0.85,
            "comment": "High quality response",
        }

        result = context.score(score)

        assert result == context
        self.mock_span.set_attribute.assert_any_call(
            "generation_score.quality.comment", "High quality response"
        )
        self.mock_span.set_attribute.assert_any_call(
            "generation_score.quality.value", "0.85"
        )
        self.mock_span.add_event.assert_called_once_with(
            "generation_score",
            {
                "name": "quality",
                "value": "0.85",
                "comment": "High quality response",
            },
        )

    def test_score_observation_context(self):
        context = ConcreteContext(self.mock_span)
        score: Score = {
            "name": "relevance",
            "value": "high",
            "comment": "Very relevant",
        }

        result = context.score(score)

        assert result == context
        self.mock_span.set_attribute.assert_any_call(
            "span_score.relevance.comment", "Very relevant"
        )
        self.mock_span.set_attribute.assert_any_call(
            "span_score.relevance.value", "high"
        )

    def test_score_with_exception_during_processing(self):
        context = ConcreteContext(self.mock_span)
        score: Score = {"name": "test", "value": 1.0, "comment": "test comment"}

        self.mock_span.set_attribute.side_effect = Exception("Attribute setting failed")

        with patch("verse_sdk.contexts.base.logging") as mock_logging:
            result = context.score(score)

            assert result == context
            mock_logging.warning.assert_called_once()

    def test_set_attributes_with_prefix(self):
        context = ConcreteContext(self.mock_span)

        result = context.set_attributes("test", key1="value1", key2="value2")

        assert result == context
        self.mock_span.set_attribute.assert_any_call("test.key1", "value1")
        self.mock_span.set_attribute.assert_any_call("test.key2", "value2")

    def test_set_attributes_without_prefix(self):
        context = ConcreteContext(self.mock_span)

        result = context.set_attributes(key1="value1", key2="value2")

        assert result == context
        self.mock_span.set_attribute.assert_any_call("key1", "value1")
        self.mock_span.set_attribute.assert_any_call("key2", "value2")

    def test_set_attributes_filters_none_values(self):
        context = ConcreteContext(self.mock_span)

        context.set_attributes("test", key1="value1", key2=None, key3="value3")

        self.mock_span.set_attribute.assert_any_call("test.key1", "value1")
        self.mock_span.set_attribute.assert_any_call("test.key3", "value3")
        assert not any(
            call[0][0] == "test.key2"
            for call in self.mock_span.set_attribute.call_args_list
        )

    def test_set_attributes_with_exception_during_processing(self):
        context = ConcreteContext(self.mock_span)

        self.mock_span.set_attribute.side_effect = Exception("Attribute setting failed")

        with patch("verse_sdk.contexts.base.logging") as mock_logging:
            result = context.set_attributes(key1="value1")

            assert result == context
            mock_logging.warning.assert_called_once()

    def test_method_chaining(self):
        context = ConcreteContext(self.mock_span)
        metadata: ContextMetadata = {"key": "value"}
        score: Score = {"name": "test", "value": 1.0, "comment": "test"}
        exception = ValueError("test error")

        result = (
            context.metadata(metadata)
            .score(score)
            .error(exception)
            .set_attributes("test", key="value")
        )

        assert result == context

    def test_get_trace_metadata_success(self):
        context = ConcreteContext(self.mock_span)
        mock_span_context = Mock()
        mock_span_context.span_id = 0x1234567890ABCDEF
        mock_span_context.trace_id = 0x1234567890ABCDEF1234567890ABCDEF
        self.mock_span.context = mock_span_context

        result = context.get_trace_metadata()

        expected_result = {
            "parent_observation_id": "1234567890abcdef",
            "trace_id": "1234567890abcdef1234567890abcdef",
        }
        assert result == expected_result

    def test_get_trace_metadata_with_zero_ids(self):
        context = ConcreteContext(self.mock_span)
        mock_span_context = Mock()
        mock_span_context.span_id = 0x0
        mock_span_context.trace_id = 0x0
        self.mock_span.context = mock_span_context

        result = context.get_trace_metadata()

        expected_result = {
            "parent_observation_id": "0000000000000000",
            "trace_id": "00000000000000000000000000000000",
        }
        assert result == expected_result

    def test_get_trace_metadata_with_max_values(self):
        context = ConcreteContext(self.mock_span)
        mock_span_context = Mock()
        mock_span_context.span_id = 0xFFFFFFFFFFFFFFFF
        mock_span_context.trace_id = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
        self.mock_span.context = mock_span_context

        result = context.get_trace_metadata()

        expected_result = {
            "parent_observation_id": "ffffffffffffffff",
            "trace_id": "ffffffffffffffffffffffffffffffff",
        }
        assert result == expected_result

    def test_get_trace_metadata_hex_formatting(self):
        context = ConcreteContext(self.mock_span)
        mock_span_context = Mock()
        mock_span_context.span_id = 0x1A2B3C4D5E6F7890
        mock_span_context.trace_id = 0x1A2B3C4D5E6F78901A2B3C4D5E6F7890
        self.mock_span.context = mock_span_context

        result = context.get_trace_metadata()

        expected_result = {
            "parent_observation_id": "1a2b3c4d5e6f7890",
            "trace_id": "1a2b3c4d5e6f78901a2b3c4d5e6f7890",
        }
        assert result == expected_result

    def test_get_trace_metadata_return_type(self):
        context = ConcreteContext(self.mock_span)

        mock_span_context = Mock()
        mock_span_context.span_id = 0x1234567890ABCDEF
        mock_span_context.trace_id = 0x1234567890ABCDEF1234567890ABCDEF
        self.mock_span.context = mock_span_context

        result = context.get_trace_metadata()

        assert isinstance(result, dict)
        assert "parent_observation_id" in result
        assert "trace_id" in result
        assert isinstance(result["parent_observation_id"], str)
        assert isinstance(result["trace_id"], str)

    def test_get_trace_metadata_string_lengths(self):
        context = ConcreteContext(self.mock_span)

        mock_span_context = Mock()
        mock_span_context.span_id = 0x1234567890ABCDEF
        mock_span_context.trace_id = 0x1234567890ABCDEF1234567890ABCDEF
        self.mock_span.context = mock_span_context

        result = context.get_trace_metadata()

        assert len(result["parent_observation_id"]) == 16
        assert len(result["trace_id"]) == 32

    def test_get_trace_metadata_hex_strings(self):
        context = ConcreteContext(self.mock_span)

        mock_span_context = Mock()
        mock_span_context.span_id = 0x1234567890ABCDEF
        mock_span_context.trace_id = 0x1234567890ABCDEF1234567890ABCDEF
        self.mock_span.context = mock_span_context

        result = context.get_trace_metadata()

        import re

        hex_pattern = re.compile(r"^[0-9a-f]+$")
        assert hex_pattern.match(result["parent_observation_id"])
        assert hex_pattern.match(result["trace_id"])

    def test_get_trace_metadata_different_contexts(self):
        """Test that different context instances return different metadata"""
        mock_span1 = Mock()
        mock_span1.get_span_context.return_value.is_valid = True
        mock_span_context1 = Mock()
        mock_span_context1.span_id = 0x1111111111111111
        mock_span_context1.trace_id = 0x11111111111111111111111111111111
        mock_span1.context = mock_span_context1

        mock_span2 = Mock()
        mock_span2.get_span_context.return_value.is_valid = True
        mock_span_context2 = Mock()
        mock_span_context2.span_id = 0x2222222222222222
        mock_span_context2.trace_id = 0x22222222222222222222222222222222
        mock_span2.context = mock_span_context2

        context1 = ConcreteContext(mock_span1)
        context2 = ConcreteContext(mock_span2)

        result1 = context1.get_trace_metadata()
        result2 = context2.get_trace_metadata()

        assert result1 != result2
        assert result1["parent_observation_id"] == "1111111111111111"
        assert result1["trace_id"] == "11111111111111111111111111111111"
        assert result2["parent_observation_id"] == "2222222222222222"
        assert result2["trace_id"] == "22222222222222222222222222222222"



import json
import logging
from typing import Any

from ...contexts import CompletionChoice, GenerationUsage


def get_finish_reason(response_obj: Any) -> str | None:
    """Extract finish reason from Anthropic response."""
    try:
        if hasattr(response_obj, "stop_reason"):
            return response_obj.stop_reason
        return None
    except Exception as e:
        logging.warning("Error getting finish reason from Anthropic", exc_info=e)
        return None


def get_completion_choices(response_obj: Any) -> list[CompletionChoice]:
    """Extract completion choices from Anthropic response."""
    try:
        choices = []
        if hasattr(response_obj, "content") and response_obj.content:
            for content_block in response_obj.content:
                if hasattr(content_block, "type") and content_block.type == "tool_use":
                    tool_call = {
                        "id": content_block.id,
                        "type": "function",
                        "function": {
                            "name": content_block.name,
                            "arguments": json.dumps(
                                content_block.input, ensure_ascii=False, default=repr
                            ),
                        },
                    }
                    choices.append(
                        CompletionChoice(
                            message={
                                "role": getattr(response_obj, "role", "assistant"),
                                "content": None,
                                "function_call": None,
                                "name": None,
                                "tool_calls": [tool_call],
                            },
                            delta=None,
                            text=None,
                        )
                    )
                elif hasattr(content_block, "text"):
                    choices.append(
                        CompletionChoice(
                            message={
                                "role": getattr(response_obj, "role", "assistant"),
                                "content": content_block.text,
                                "function_call": None,
                                "name": None,
                                "tool_calls": None,
                            },
                            delta=None,
                            text=content_block.text,
                        )
                    )
        return choices
    except Exception as e:
        logging.warning("Error getting completion choices from Anthropic", exc_info=e)
        return []


def get_usage_from_response(response_obj: Any) -> GenerationUsage:
    """Extract usage information from Anthropic response."""
    try:
        usage = getattr(response_obj, "usage", None)
        if usage:
            input_tokens = getattr(usage, "input_tokens", None)
            output_tokens = getattr(usage, "output_tokens", None)
            total_tokens = None
            if input_tokens is not None and output_tokens is not None:
                total_tokens = input_tokens + output_tokens

            return {
                "input_tokens": input_tokens,
                "output_tokens": output_tokens,
                "prompt_tokens": input_tokens,
                "completion_tokens": output_tokens,
                "total_tokens": total_tokens,
                "max_tokens": None,
                "temperature": None,
                "top_p": None,
                "stream": None,
                "tool_calls": None,
            }
        return {}
    except Exception as e:
        logging.warning("Error getting usage from Anthropic response", exc_info=e)
        return {}


def get_request_parameters(kwargs: dict) -> GenerationUsage:
    """Extract request parameters from kwargs."""
    try:
        return {
            "max_tokens": kwargs.get("max_tokens"),
            "temperature": kwargs.get("temperature"),
            "top_p": kwargs.get("top_p"),
            "stream": kwargs.get("stream", False),
            "input_tokens": None,
            "output_tokens": None,
            "prompt_tokens": None,
            "completion_tokens": None,
            "total_tokens": None,
            "tool_calls": None,
        }
    except Exception as e:
        logging.warning("Error getting request parameters from Anthropic", exc_info=e)
        return {}


def format_messages_for_tracing(messages: list[dict]) -> list[dict]:
    """Format Anthropic messages for tracing."""
    try:
        formatted_messages = []
        for msg in messages:
            role = msg.get("role", "user")
            content = msg.get("content", "")

            if isinstance(content, list):
                text_content = []
                for block in content:
                    if isinstance(block, dict):
                        if block.get("type") == "text":
                            text_content.append(block.get("text", ""))
                        elif block.get("type") == "tool_result":
                            text_content.append(f"[Tool Result: {block.get('content', '')}]")
                    elif hasattr(block, "type"):
                        if block.type == "text":
                            text_content.append(getattr(block, "text", ""))
                        elif block.type == "tool_result":
                            text_content.append(f"[Tool Result: {getattr(block, 'content', '')}]")
                content = "\n".join(text_content)

            formatted_messages.append({
                "role": role,
                "content": str(content),
                "function_call": None,
                "name": None,
                "tool_calls": None,
            })
        return formatted_messages
    except Exception as e:
        logging.warning("Error formatting messages for tracing", exc_info=e)
        return []


def get_stream_text(chunk: Any) -> str:
    """Extract text from streaming chunk."""
    try:
        if hasattr(chunk, "delta"):
            delta = chunk.delta
            if hasattr(delta, "text"):
                return delta.text or ""
        if hasattr(chunk, "content_block"):
            content_block = chunk.content_block
            if hasattr(content_block, "text"):
                return content_block.text or ""
        return ""
    except Exception:
        return ""



from collections.abc import Iterator
from typing import ClassVar

from .types import OPTIONAL, RECOMMENDED, REQUIRED, AttributeSpec


class OtelGenAISpanSpecification:
    SPEC_VERSION: ClassVar[str] = "1.38.0"

    SUPPORTED_ATTRIBUTES: ClassVar[dict[str, AttributeSpec]] = {
        "aws.bedrock.guardrail.id": AttributeSpec(
            name="aws.bedrock.guardrail.id",
            requirement=REQUIRED,
            type="string",
        ),
        "aws.bedrock.knowledge_base.id": AttributeSpec(
            name="aws.bedrock.knowledge_base.id",
            requirement=RECOMMENDED,
            type="string",
        ),
        "error.type": AttributeSpec(
            name="error.type",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.agent.description": AttributeSpec(
            name="gen_ai.agent.description",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.agent.id": AttributeSpec(
            name="gen_ai.agent.id",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.agent.name": AttributeSpec(
            name="gen_ai.agent.name",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.completion": AttributeSpec(
            deprecated=True,
            deprecated_message="Deprecated in v1.36.0+. Use Event API with gen_ai.output.messages instead",
            name="gen_ai.completion",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.content.completion": AttributeSpec(
            deprecated=True,
            deprecated_message="Deprecated in v1.36.0+. Use Event API with gen_ai.output.messages instead",
            name="gen_ai.content.completion",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.content.prompt": AttributeSpec(
            deprecated=True,
            deprecated_message="Deprecated in v1.36.0+. Use Event API with gen_ai.input.messages instead",
            name="gen_ai.content.prompt",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.conversation.id": AttributeSpec(
            name="gen_ai.conversation.id",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.data_source.id": AttributeSpec(
            name="gen_ai.data_source.id",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.embeddings.dimension.count": AttributeSpec(
            name="gen_ai.embeddings.dimension.count",
            requirement=RECOMMENDED,
            type="int",
        ),
        # Event-specific attributes
        #
        # # "gen_ai.evaluation.explanation": AttributeSpec(
        #     name="gen_ai.evaluation.explanation",
        #     requirement=RECOMMENDED,
        #     type="string",
        # ),
        # "gen_ai.evaluation.name": AttributeSpec(
        #     name="gen_ai.evaluation.name",
        #     requirement=REQUIRED,
        #     type="string",
        # ),
        # "gen_ai.evaluation.score.label": AttributeSpec(
        #     name="gen_ai.evaluation.score.label",
        #     requirement=RECOMMENDED,
        #     type="string",
        # ),
        # "gen_ai.evaluation.score.value": AttributeSpec(
        #     name="gen_ai.evaluation.score.value",
        #     requirement=RECOMMENDED,
        #     type="double",
        # ),
        "gen_ai.input.messages": AttributeSpec(
            name="gen_ai.input.messages",
            requirement=OPTIONAL,
            type="any",
        ),
        "gen_ai.operation.name": AttributeSpec(
            name="gen_ai.operation.name",
            requirement=REQUIRED,
            type="string",
            valid_values=[
                "chat",
                "create_agent",
                "embeddings",
                "execute_tool",
                "generate_content",
                "invoke_agent",
                "text_completion",
            ],
        ),
        "gen_ai.output.messages": AttributeSpec(
            name="gen_ai.output.messages",
            requirement=OPTIONAL,
            type="any",
        ),
        "gen_ai.output.type": AttributeSpec(
            name="gen_ai.output.type",
            requirement=OPTIONAL,
            type="string",
            valid_values=[
                "image",
                "json",
                "speech",
                "text",
            ],
        ),
        "gen_ai.prompt": AttributeSpec(
            deprecated=True,
            deprecated_message="Deprecated in v1.36.0+. Use Event API with gen_ai.input.messages instead",
            name="gen_ai.prompt",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.provider.name": AttributeSpec(
            name="gen_ai.provider.name",
            requirement=REQUIRED,
            type="string",
            valid_values=[
                "anthropic",
                "aws.bedrock",
                "azure.ai.inference",
                "azure.ai.openai",
                "cohere",
                "deepseek",
                "gcp.gemini",
                "gcp.gen_ai",
                "gcp.vertex_ai",
                "groq",
                "ibm.watsonx.ai",
                "mistral_ai",
                "openai",
                "perplexity",
                "x_ai",
            ],
        ),
        "gen_ai.request.choice.count": AttributeSpec(
            name="gen_ai.request.choice.count",
            requirement=OPTIONAL,
            type="int",
        ),
        "gen_ai.request.encoding_formats": AttributeSpec(
            name="gen_ai.request.encoding_formats",
            requirement=RECOMMENDED,
            type="string[]",
        ),
        "gen_ai.request.frequency_penalty": AttributeSpec(
            name="gen_ai.request.frequency_penalty",
            requirement=RECOMMENDED,
            type="double",
        ),
        "gen_ai.request.max_tokens": AttributeSpec(
            name="gen_ai.request.max_tokens",
            requirement=RECOMMENDED,
            type="int",
        ),
        "gen_ai.request.model": AttributeSpec(
            name="gen_ai.request.model",
            requirement=OPTIONAL,
            type="string",
        ),
        "gen_ai.request.presence_penalty": AttributeSpec(
            name="gen_ai.request.presence_penalty",
            requirement=RECOMMENDED,
            type="double",
        ),
        "gen_ai.request.seed": AttributeSpec(
            name="gen_ai.request.seed",
            requirement=OPTIONAL,
            type="int",
        ),
        "gen_ai.request.stop_sequences": AttributeSpec(
            name="gen_ai.request.stop_sequences",
            requirement=RECOMMENDED,
            type="string[]",
        ),
        "gen_ai.request.temperature": AttributeSpec(
            name="gen_ai.request.temperature",
            requirement=RECOMMENDED,
            type="double",
        ),
        "gen_ai.request.top_k": AttributeSpec(
            name="gen_ai.request.top_k",
            requirement=RECOMMENDED,
            type="double",
        ),
        "gen_ai.request.top_p": AttributeSpec(
            name="gen_ai.request.top_p",
            requirement=RECOMMENDED,
            type="double",
        ),
        "gen_ai.response.finish_reasons": AttributeSpec(
            name="gen_ai.response.finish_reasons",
            requirement=RECOMMENDED,
            type="string[]",
            valid_values=[
                "content_filter",
                "error",
                "length",
                "other",
                "stop",
                "tool_calls",
            ],
        ),
        "gen_ai.response.id": AttributeSpec(
            name="gen_ai.response.id",
            requirement=RECOMMENDED,
            type="string",
        ),
        "gen_ai.response.model": AttributeSpec(
            name="gen_ai.response.model",
            requirement=RECOMMENDED,
            type="string",
        ),
        "gen_ai.system": AttributeSpec(
            deprecated=True,
            deprecated_message="Deprecated in v1.38.0+. Use gen_ai.provider.name instead",
            name="gen_ai.system",
            requirement=OPTIONAL,
            type="string",
            valid_values=[
                "anthropic",
                "azure_ai",
                "cohere",
                "openai",
                "vertex_ai",
            ],
        ),
        "gen_ai.system_instructions": AttributeSpec(
            name="gen_ai.system_instructions",
            requirement=OPTIONAL,
            type="any",
        ),
        # Event-specific attribute
        #
        # "gen_ai.token.type": AttributeSpec(
        #     name="gen_ai.token.type",
        #     requirement=OPTIONAL,
        #     type="string",
        #     valid_values=[
        #         "input",
        #         "output",
        #     ],
        # ),
        "gen_ai.tool.call.arguments": AttributeSpec(
            name="gen_ai.tool.call.arguments",
            requirement=OPTIONAL,
            type="any",
        ),
        "gen_ai.tool.call.id": AttributeSpec(
            name="gen_ai.tool.call.id",
            requirement=RECOMMENDED,
            type="string",
        ),
        "gen_ai.tool.call.result": AttributeSpec(
            name="gen_ai.tool.call.result",
            requirement=OPTIONAL,
            type="any",
        ),
        "gen_ai.tool.definitions": AttributeSpec(
            name="gen_ai.tool.definitions",
            requirement=OPTIONAL,
            type="any",
        ),
        "gen_ai.tool.description": AttributeSpec(
            name="gen_ai.tool.description",
            requirement=RECOMMENDED,
            type="string",
        ),
        "gen_ai.tool.name": AttributeSpec(
            name="gen_ai.tool.name",
            requirement=RECOMMENDED,
            type="string",
        ),
        "gen_ai.tool.type": AttributeSpec(
            name="gen_ai.tool.type",
            requirement=RECOMMENDED,
            type="string",
            valid_values=[
                "datastore",
                "extension",
                "function",
            ],
        ),
        "gen_ai.usage.input_tokens": AttributeSpec(
            name="gen_ai.usage.input_tokens",
            requirement=RECOMMENDED,
            type="int",
        ),
        "gen_ai.usage.output_tokens": AttributeSpec(
            name="gen_ai.usage.output_tokens",
            requirement=RECOMMENDED,
            type="int",
        ),
        "openai.request.service_tier": AttributeSpec(
            name="openai.request.service_tier",
            requirement=OPTIONAL,
            type="string",
            valid_values=[
                "auto",
                "default",
            ],
        ),
        "openai.response.service_tier": AttributeSpec(
            name="openai.response.service_tier",
            requirement=RECOMMENDED,
            type="string",
            valid_values=[
                "default",
                "scale",
            ],
        ),
        "openai.response.system_fingerprint": AttributeSpec(
            name="openai.response.system_fingerprint",
            requirement=RECOMMENDED,
            type="string",
        ),
        "server.address": AttributeSpec(
            name="server.address",
            requirement=RECOMMENDED,
            type="string",
        ),
        "server.port": AttributeSpec(
            name="server.port",
            requirement=OPTIONAL,
            type="int",
        ),
    }

    def __iter__(self) -> Iterator[AttributeSpec]:
        return iter(self.SUPPORTED_ATTRIBUTES.values())



import functools
import logging
from typing import Any

from opentelemetry.trace import Status, StatusCode

from ...contexts import GenerationContext
from ...sdk import VerseSDK
from .utils import (
    format_messages_for_tracing,
    get_completion_choices,
    get_finish_reason,
    get_request_parameters,
    get_stream_text,
    get_usage_from_response,
)


def configure_google(sdk: VerseSDK):
    """Configure Google GenAI to use the Verse SDK."""
    try:
        from google.genai.models import Models
    except ImportError:
        logging.warning("Google GenAI SDK not installed, skipping integration")
        return

    original_generate_content = Models.generate_content
    original_generate_content_stream = Models.generate_content_stream

    @functools.wraps(original_generate_content)
    def generate_content_with_tracing(self, *args, **kwargs):
        """Wrap synchronous generate_content with tracing."""
        model_name = kwargs.get("model", "unknown")
        contents = kwargs.get("contents", [])

        generation_ctx = sdk.create_generation(
            name="Google: Generate Content",
            model=model_name,
            usage=get_request_parameters(kwargs),
        )

        generation_ctx.provider("gcp.gemini").operation("generate_content").request_params(
            get_request_parameters(kwargs)
        )

        if contents:
            formatted_messages = format_messages_for_tracing(contents)
            generation_ctx.messages(formatted_messages)

        try:
            response = original_generate_content(self, *args, **kwargs)
            _process_response(response, generation_ctx, model_name)
            generation_ctx.span.end()
            return response

        except Exception as e:
            generation_ctx.error(e)
            generation_ctx.span.end()
            raise

    @functools.wraps(original_generate_content_stream)
    def generate_content_stream_with_tracing(self, *args, **kwargs):
        """Wrap streaming generate_content with tracing."""
        model_name = kwargs.get("model", "unknown")
        contents = kwargs.get("contents", [])

        generation_ctx = sdk.create_generation(
            name="Google: Generate Content Stream",
            model=model_name,
            usage=get_request_parameters(kwargs),
        )

        generation_ctx.provider("gcp.gemini").operation("generate_content").request_params(
            get_request_parameters(kwargs)
        )

        if contents:
            formatted_messages = format_messages_for_tracing(contents)
            generation_ctx.messages(formatted_messages)

        try:
            stream = original_generate_content_stream(self, *args, **kwargs)
            return _wrap_stream_response(stream, generation_ctx, sdk)

        except Exception as e:
            generation_ctx.error(e)
            generation_ctx.span.end()
            raise

    Models.generate_content = generate_content_with_tracing
    Models.generate_content_stream = generate_content_stream_with_tracing


def _process_response(response: Any, generation_ctx: GenerationContext, model_name: str):
    """Process non-streaming response and add attributes to span."""
    try:
        generation_ctx.model_used(model_name)

        finish_reason = get_finish_reason(response)
        if finish_reason:
            generation_ctx.reason(finish_reason)

        choices = get_completion_choices(response)
        if choices:
            generation_ctx.completions(choices)

        usage = get_usage_from_response(response)
        if usage:
            generation_ctx.usage(usage)

        generation_ctx.status(Status(StatusCode.OK))

    except Exception as e:
        logging.warning("Error processing Google GenAI response", exc_info=e)


def _wrap_stream_response(stream, generation_ctx: GenerationContext, sdk: VerseSDK):
    """Wrap streaming response to capture events."""
    try:
        for chunk in stream:
            text = get_stream_text(chunk)
            if text:
                sdk.event(
                    "gen_ai.stream.delta",
                    metadata={"content": text[:512]},
                    level="info",
                    span=generation_ctx.span,
                )

            yield chunk

        if hasattr(stream, "_result"):
            _process_response(stream._result, generation_ctx, generation_ctx.span.name)
        generation_ctx.span.end()

    except Exception as e:
        generation_ctx.error(e)
        generation_ctx.span.end()
        raise


async def _wrap_async_stream_response(
    stream, generation_ctx: GenerationContext, sdk: VerseSDK
):
    """Wrap async streaming response to capture events."""
    try:
        async for chunk in stream:
            text = get_stream_text(chunk)
            if text:
                sdk.event(
                    "gen_ai.stream.delta",
                    metadata={"content": text[:512]},
                    level="info",
                    span=generation_ctx.span,
                )

            yield chunk

        if hasattr(stream, "_result"):
            _process_response(stream._result, generation_ctx, generation_ctx.span.name)
        generation_ctx.span.end()

    except Exception as e:
        generation_ctx.error(e)
        generation_ctx.span.end()
        raise



from __future__ import annotations

from .base import BaseContext
from .generation import GenerationContext
from .span import SpanContext
from .trace import TraceContext
from .types import (
    CompletionChoice,
    ContextMetadata,
    ContextType,
    EventMetadata,
    GenerationMessage,
    GenerationUsage,
    ObservationType,
    OperationType,
    Score,
)
from .utils import create_span_in_existing_trace, get_current_context

get_current_generation_context = get_current_context(GenerationContext)
get_current_span_context = get_current_context(SpanContext)
get_current_trace_context = get_current_context(TraceContext)


__all__ = [
    "BaseContext",
    "CompletionChoice",
    "ContextMetadata",
    "ContextType",
    "EventMetadata",
    "GenerationContext",
    "GenerationMessage",
    "GenerationUsage",
    "ObservationType",
    "OperationType",
    "Score",
    "SpanContext",
    "TraceContext",
    "create_span_in_existing_trace",
    "get_current_generation_context",
    "get_current_span_context",
    "get_current_trace_context",
]



from collections.abc import AsyncIterator
from typing import Any

from tests.fixtures.agents.prompt import SYSTEM_PROMPT
from tests.fixtures.datasource import datasource

from .litellm import run_completion, run_completion_async


class LitellmAgent:
    def ask(self, question: str) -> list[Any]:
        response = run_completion(
            [
                {"role": "system", "content": SYSTEM_PROMPT},
                {"role": "user", "content": question},
            ],
            datasource,
        )

        return response.choices

    async def stream(self, question: str) -> AsyncIterator[str]:
        async for part in run_completion_async(
            [
                {"role": "system", "content": SYSTEM_PROMPT},
                {"role": "user", "content": question},
            ],
            datasource,
        ):
            yield part



from typing import Any


def is_valid_type(value: Any, expected_type: str) -> bool:
    """Check if a value matches the expected type.

    Args:
        value: The value to check.
        expected_type: The expected type (string, int, double, boolean, string[], any).

    Returns:
        bool: True if the value matches the expected type, False otherwise.
    """
    if expected_type == "any":
        return True

    if expected_type == "string":
        return isinstance(value, str)

    if expected_type == "int":
        return isinstance(value, int) and not isinstance(value, bool)

    if expected_type == "double":
        return isinstance(value, (int, float)) and not isinstance(value, bool)

    if expected_type == "boolean":
        return isinstance(value, bool)

    if expected_type == "string[]":
        return isinstance(value, list) and all(isinstance(item, str) for item in value)

    return False



from __future__ import annotations

import logging
from typing import Any

from ..utils import get
from .span import SpanContext
from .types import (
    CompletionChoice,
    ContextType,
    GenerationMessage,
    GenerationUsage,
    ObservationType,
)
from .utils import (
    get_choice_message,
    get_content_from_choice_message,
    get_content_from_generation_message,
    get_function_call,
    get_tool_call,
    get_tool_call_from_choice_message,
    normalize_finish_reason,
)


class GenerationContext(SpanContext):
    """Context for generation spans (LLM operations)."""

    observation_type: ObservationType = "observation"
    type: ContextType = "generation"

    def __init__(self, *args, **kwargs):
        """Initialize the generation context with a special attribute."""
        super().__init__(*args, **kwargs)
        self.operation(self.observation_type)

    def completions(
        self,
        completion_choices: list[CompletionChoice],
    ) -> GenerationContext:
        """Recursively trace completion choices w/ support for tool calling."""
        try:
            for choice_index, choice in enumerate(completion_choices):
                if choice is None:
                    continue

                message = get_choice_message(choice)
                text = get(choice, "text", "")

                if message:
                    func_args, func_name = get_function_call(message)

                    attrs = {
                        f"gen_ai.completion.{choice_index}.content": str(
                            get_content_from_choice_message(message)
                        ),
                        f"gen_ai.completion.{choice_index}.function_call.name": func_name,
                        f"gen_ai.completion.{choice_index}.function_call.arguments": func_args,
                        f"gen_ai.completion.{choice_index}.role": get(
                            message, "role", "assistant"
                        ),
                    }

                    for tool_index, tool_call in enumerate(
                        get_tool_call_from_choice_message(message)
                    ):
                        tool_id, tool_type = get_tool_call(tool_call)
                        tool_func_args, tool_func_name = get_function_call(
                            tool_call, "function"
                        )

                        attrs.update(
                            {
                                f"gen_ai.completion.{choice_index}.tool_call.{tool_index}.id": tool_id,
                                f"gen_ai.completion.{choice_index}.tool_call.{tool_index}.type": tool_type,
                                f"gen_ai.completion.{choice_index}.tool_call.{tool_index}.function.name": tool_func_name,
                                f"gen_ai.completion.{choice_index}.tool_call.{tool_index}.function.arguments": tool_func_args,
                            }
                        )

                    self.set_attributes(**attrs)
                else:
                    self.set_attributes(
                        **{
                            f"gen_ai.completion.{choice_index}.role": "assistant",
                            f"gen_ai.completion.{choice_index}.content": text or "",
                        }
                    )

            return self
        except Exception as e:
            logging.warning("Completions could not be traced", exc_info=e)
            return self

    def input(self, content: Any) -> GenerationContext:
        """Set the input for the generation."""
        self._span.set_attribute("gen_ai.prompt", str(content))
        return self

    def messages(self, messages: list[GenerationMessage]) -> GenerationContext:
        """Recursively trace messages w/ support for tool calling."""
        try:
            for message_index, message in enumerate(messages):
                if not message:
                    continue

                func_args, func_name = get_function_call(message)
                attrs = {
                    f"gen_ai.prompt.{message_index}.id": get(message, "id", None),
                    f"gen_ai.prompt.{message_index}.role": get(message, "role", "user"),
                    f"gen_ai.prompt.{message_index}.content": get_content_from_generation_message(
                        message
                    ),
                    f"gen_ai.prompt.{message_index}.function_call.name": func_name,
                    f"gen_ai.prompt.{message_index}.function_call.arguments": func_args,
                    f"gen_ai.prompt.{message_index}.name": get(message, "name", None),
                }

                for tool_index, tool_call in enumerate(
                    # note: same structure as completion choices
                    get_tool_call_from_choice_message(message)
                ):
                    tool_id, tool_type = get_tool_call(tool_call)
                    tool_func_args, tool_func_name = get_function_call(
                        tool_call, "function"
                    )

                    attrs.update(
                        {
                            f"gen_ai.prompt.{message_index}.tool_call.{tool_index}.id": tool_id,
                            f"gen_ai.prompt.{message_index}.tool_call.{tool_index}.type": tool_type,
                            f"gen_ai.prompt.{message_index}.tool_call.{tool_index}.function.name": tool_func_name,
                            f"gen_ai.prompt.{message_index}.tool_call.{tool_index}.function.arguments": tool_func_args,
                        }
                    )

                self.set_attributes(**attrs)

            return self
        except Exception as e:
            logging.warning("Messages could not be traced", exc_info=e)
            return self

    def model(self, model_name: str) -> GenerationContext:
        """Set the desired model for generation."""
        self._span.set_attribute("gen_ai.request.model", model_name)
        return self

    def model_used(self, model_used: str) -> GenerationContext:
        """Set the actual model used for generation."""
        self._span.set_attribute("gen_ai.response.model", model_used)
        return self

    def operation(self, operation_name: str) -> GenerationContext:
        """Set the operation name for the generation."""
        self._span.set_attribute("gen_ai.operation.name", operation_name)
        return self

    def output(self, content: Any) -> GenerationContext:
        """Set the output of the generation."""
        self._span.set_attribute("gen_ai.response.completion", str(content))
        return self

    def provider(self, provider_name: str) -> GenerationContext:
        """Set the provider name for the generation."""
        self._span.set_attribute("gen_ai.provider.name", provider_name)
        return self

    def reason(self, reason: str) -> GenerationContext:
        """Set the reason for the generation."""
        normalized = normalize_finish_reason(reason)
        if normalized:
            self._span.set_attribute("gen_ai.response.finish_reasons", [normalized])
        return self

    def request_params(self, usage: GenerationUsage) -> GenerationContext:
        """Set request parameters from usage dict."""
        if usage.get("max_tokens") is not None:
            self._span.set_attribute("gen_ai.request.max_tokens", usage["max_tokens"])
        if usage.get("temperature") is not None:
            self._span.set_attribute("gen_ai.request.temperature", usage["temperature"])
        if usage.get("top_p") is not None:
            self._span.set_attribute("gen_ai.request.top_p", usage["top_p"])
        if usage.get("top_k") is not None:
            self._span.set_attribute("gen_ai.request.top_k", usage["top_k"])
        return self

    def response_id(self, response_id: str) -> GenerationContext:
        """Set the response ID for the generation."""
        self._span.set_attribute("gen_ai.response.id", response_id)
        return self

    def usage(
        self,
        usage: GenerationUsage,
    ) -> GenerationContext:
        """Set token usage metrics."""
        self.set_attributes(
            prefix="gen_ai.usage",
            **usage,
        )

        return self

    def vendor(self, vendor: str) -> GenerationContext:
        """Set the vendor for the generation."""
        self._span.set_attribute("gen_ai.request.vendor", vendor)
        return self



LANGFUSE_ENDPOINT = "/api/public/otel/v1/traces"
LANGFUSE_HOST = "LANGFUSE_HOST"
LANGFUSE_PUBLIC_KEY = "LANGFUSE_PUBLIC_KEY"
LANGFUSE_REGION = "LANGFUSE_REGION"
LANGFUSE_REGION_US = "US"
LANGFUSE_SECRET_KEY = "LANGFUSE_SECRET_KEY"
LANGFUSE_URL_EU = f"https://cloud.langfuse.com{LANGFUSE_ENDPOINT}"
LANGFUSE_URL_US = f"https://us.cloud.langfuse.com{LANGFUSE_ENDPOINT}"



from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import SpanProcessor
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor

from ...wrappers import ScopeFilterExporter
from ..base import Exporter
from ..types import ExporterConfig


class ConsoleExporter(Exporter):
    config: ExporterConfig

    def __init__(self, config: ExporterConfig):
        self.config = config

    def create_span_processor(self, resource: Resource) -> SpanProcessor:
        """Create a span processor for the console exporter w/ scope filtering."""
        exporter = ScopeFilterExporter(
            ConsoleSpanExporter(),
            self.config,
        )

        return SimpleSpanProcessor(exporter)



from __future__ import annotations

import logging
from contextlib import suppress
from contextvars import ContextVar
from typing import Any, ClassVar

from opentelemetry.trace import Status, StatusCode

from ...contexts import GenerationContext
from ...sdk import VerseSDK

_VERSE_LANGCHAIN_HANDLER_VAR: ContextVar[Any] = ContextVar(
    "verse_langchain_handler", default=None
)
_VERSE_LANGCHAIN_HOOK_REGISTERED = False


def configure_langchain(sdk: VerseSDK) -> None:
    """
    Register the Verse handler with LangChain via the configure hook mechanism so
    instrumentation is applied automatically to all chains.
    """
    from langchain_core.callbacks.base import BaseCallbackHandler

    from .utils import (
        get_completions_from_result,
        get_generation_messages_from_chat,
        get_generation_messages_from_prompts,
        get_usage_from_result,
        get_vendor_and_model,
    )

    class VerseLangChainHandler(BaseCallbackHandler):
        """
        Minimal LangChain callback handler that records GenAI metadata.
        """

        current: ClassVar[dict[Any, GenerationContext]] = {}

        def __init__(self, sdk: VerseSDK):
            super().__init__()
            self._sdk = sdk

        def on_chat_model_start(self, serialized: Any, messages, run_id, **kwargs):
            vendor, model = get_vendor_and_model(serialized, **kwargs)
            prompts = get_generation_messages_from_chat(messages)
            ctx = self._sdk.create_generation(
                name="LangChain: Chat",
                model=model,
                messages=prompts,
                vendor=vendor,
            )
            ctx.provider(vendor).operation("chat")
            self.current[run_id] = ctx

        def on_chat_model_end(self, result, run_id, **kwargs):
            self._finish_generation(result, run_id)

        def on_chat_model_error(self, error: Exception, run_id, **kwargs):
            self._handle_error(error, run_id)

        def on_llm_start(self, serialized: Any, prompts, run_id, **kwargs):
            vendor, model = get_vendor_and_model(serialized, **kwargs)
            prompt_messages = get_generation_messages_from_prompts(prompts)
            ctx = self._sdk.create_generation(
                name="LangChain: Completion",
                model=model,
                messages=prompt_messages,
                vendor=vendor,
            )
            ctx.provider(vendor).operation("chat")
            self.current[run_id] = ctx

        def on_llm_end(self, result, run_id, **kwargs):
            self._finish_generation(result, run_id)

        def on_llm_new_token(self, token: str, run_id, **kwargs):
            ctx = self.current.get(run_id)
            if ctx and token:
                with suppress(Exception):
                    ctx.event("gen_ai.stream.delta", metadata={"content": token[:512]})

        def on_llm_error(self, error: Exception, run_id, **kwargs):
            self._handle_error(error, run_id)

        def on_embedding_start(self, serialized: Any, inputs, run_id, **kwargs):
            vendor, model = get_vendor_and_model(serialized, **kwargs)
            ctx = self._sdk.create_generation(
                name="LangChain: Embedding",
                model=model,
                vendor=vendor,
            )
            ctx.provider(vendor).operation("embedding")
            self.current[run_id] = ctx

        def on_embedding_end(self, result, run_id, **kwargs):
            ctx = self.current.pop(run_id, None)
            if not ctx:
                return
            try:
                ctx.status(Status(StatusCode.OK))
            except Exception as exc:
                ctx.error(exc)
            finally:
                ctx.span.end()

        def on_embedding_error(self, error: Exception, run_id, **kwargs):
            self._handle_error(error, run_id)

        def _finish_generation(self, result: Any, run_id: Any) -> None:
            ctx = self.current.pop(run_id, None)
            if not ctx:
                return

            try:
                completions, finish_reason, model_used = get_completions_from_result(
                    result
                )

                if completions:
                    ctx.completions(completions)

                if finish_reason:
                    ctx.reason(finish_reason)

                usage = get_usage_from_result(result)
                if any(value is not None for value in usage.values()):
                    ctx.usage(usage)

                if model_used:
                    ctx.model_used(model_used)

                ctx.status(Status(StatusCode.OK))
            except Exception as exc:
                ctx.error(exc)
            finally:
                ctx.span.end()

        def _handle_error(self, error: Exception, run_id: Any) -> None:
            ctx = self.current.pop(run_id, None)
            if not ctx:
                return
            ctx.error(error)
            ctx.span.end()

    handler = getattr(sdk, "_verse_langchain_handler", None)
    if handler is None:
        handler = VerseLangChainHandler(sdk)
        sdk._verse_langchain_handler = handler
    else:
        handler._sdk = sdk

    _VERSE_LANGCHAIN_HANDLER_VAR.set(handler)

    from langchain_core.tracers.context import register_configure_hook

    global _VERSE_LANGCHAIN_HOOK_REGISTERED

    if not _VERSE_LANGCHAIN_HOOK_REGISTERED:
        register_configure_hook(_VERSE_LANGCHAIN_HANDLER_VAR, inheritable=True)
        _VERSE_LANGCHAIN_HOOK_REGISTERED = True

    logging.info("LangChain handler registered globally via configure hook")



"""Tests for LangChain integration utilities."""

import types

from verse_sdk.integrations.langchain.utils import (
    get_completions_from_result,
    get_generation_messages_from_chat,
    get_generation_messages_from_prompts,
    get_usage_from_result,
    get_vendor_and_model,
)


def make_chat_message(class_name, content, **extra):
    message_cls = type(class_name, (), {})
    message = message_cls()
    message.content = content
    for key, value in extra.items():
        setattr(message, key, value)
    return message


def make_chat_generation(content, *, tool_calls=None, finish_reason="stop"):
    message = make_chat_message(
        "AIMessage",
        content,
        additional_kwargs={"tool_calls": tool_calls} if tool_calls else {},
        response_metadata={"finish_reason": finish_reason},
    )
    return types.SimpleNamespace(
        message=message,
        generation_info={"finish_reason": finish_reason},
    )


def make_text_generation(text, finish_reason="stop"):
    return types.SimpleNamespace(
        text=text,
        generation_info={"finish_reason": finish_reason},
        message=None,
    )


def make_llm_result(
    generations, *, model_name="gpt-4o", prompt=10, completion=5, total=15
):
    return types.SimpleNamespace(
        generations=generations,
        llm_output={
            "model_name": model_name,
            "token_usage": {
                "prompt_tokens": prompt,
                "completion_tokens": completion,
                "total_tokens": total,
            },
        },
    )


class TestGetVendorAndModel:
    def test_extracts_vendor_from_serialized_id(self):
        vendor, model = get_vendor_and_model(
            {"id": ["langchain", "openai", "chat"]},
            invocation_params={"model": "gpt-4"},
        )
        assert vendor == "openai"
        assert model == "gpt-4"

    def test_returns_unknown_for_unrecognized_vendor(self):
        vendor, model = get_vendor_and_model(
            {"id": ["langchain", "custom", "llm"]},
        )
        assert vendor == "unknown"
        assert model == "unknown"

    def test_extracts_model_from_kwargs(self):
        _vendor, model = get_vendor_and_model(
            {"id": ["langchain", "openai"]}, model="gpt-3.5-turbo"
        )
        assert model == "gpt-3.5-turbo"


class TestGetGenerationMessagesFromChat:
    def test_converts_chat_messages_to_generation_format(self):
        sys_msg = make_chat_message("SystemMessage", "You are helpful")
        user_msg = make_chat_message("HumanMessage", "Hi")

        messages = get_generation_messages_from_chat([[sys_msg, user_msg]])

        assert len(messages) == 2
        assert messages[0]["role"] == "system"
        assert messages[0]["content"] == "You are helpful"
        assert messages[1]["role"] == "user"
        assert messages[1]["content"] == "Hi"

    def test_handles_empty_message_batches(self):
        messages = get_generation_messages_from_chat([])
        assert messages == []

    def test_uses_explicit_role_when_available(self):
        assistant_msg = make_chat_message("ChatMessage", "Hello!", role="assistant")

        messages = get_generation_messages_from_chat([[assistant_msg]])

        assert messages == [
            {
                "role": "assistant",
                "content": "Hello!",
                "function_call": None,
                "name": None,
                "tool_calls": None,
            }
        ]


class TestGetGenerationMessagesFromPrompts:
    def test_converts_prompts_to_user_messages(self):
        messages = get_generation_messages_from_prompts(["prompt1", "prompt2"])

        assert len(messages) == 2
        assert all(m["role"] == "user" for m in messages)
        assert messages[0]["content"] == "prompt1"
        assert messages[1]["content"] == "prompt2"

    def test_handles_empty_prompts(self):
        messages = get_generation_messages_from_prompts([])
        assert messages == []


class TestGetCompletionsFromResult:
    def test_extracts_chat_completions_with_tool_calls(self):
        tool_calls = [{"type": "function", "function": {"name": "weather"}}]
        result = make_llm_result(
            [[make_chat_generation("response", tool_calls=tool_calls)]]
        )

        completions, finish_reason, model_used = get_completions_from_result(result)

        assert len(completions) == 1
        assert completions[0]["message"]["content"] == "response"
        assert completions[0]["message"]["tool_calls"] == tool_calls
        assert finish_reason == "stop"
        assert model_used == "gpt-4o"

    def test_extracts_text_completions(self):
        result = make_llm_result([[make_text_generation("A haiku appears.")]])

        completions, finish_reason, _model_used = get_completions_from_result(result)

        assert len(completions) == 1
        assert completions[0]["text"] == "A haiku appears."
        assert completions[0]["message"] is None
        assert finish_reason == "stop"

    def test_handles_multiple_generations(self):
        result = make_llm_result(
            [[make_text_generation("first"), make_text_generation("second")]]
        )

        completions, _, _ = get_completions_from_result(result)
        assert len(completions) == 2


class TestGetUsageFromResult:
    def test_extracts_token_usage(self):
        result = make_llm_result([[]], prompt=100, completion=50, total=150)
        usage = get_usage_from_result(result)

        assert usage["prompt_tokens"] == 100
        assert usage["completion_tokens"] == 50
        assert usage["total_tokens"] == 150
        assert usage["input_tokens"] == 100
        assert usage["output_tokens"] == 50

    def test_calculates_total_when_missing(self):
        result = types.SimpleNamespace(
            llm_output={
                "token_usage": {
                    "prompt_tokens": 30,
                    "completion_tokens": 20,
                }
            }
        )
        usage = get_usage_from_result(result)
        assert usage["total_tokens"] == 50

    def test_returns_empty_usage_when_no_data(self):
        result = types.SimpleNamespace(llm_output={})
        usage = get_usage_from_result(result)

        assert usage["prompt_tokens"] is None
        assert usage["completion_tokens"] is None
        assert usage["total_tokens"] is None



import json

from verse_sdk import observe, verse


def parse_spans(out: str) -> list[dict]:
    """Parse the spans from the output."""
    spans = []
    buffer = ""
    for line in out.splitlines():
        line = line.strip()
        if not line:
            continue
        buffer += line
        if buffer.count("{") == buffer.count("}"):
            try:
                span = json.loads(buffer)
                spans.append(span)
            except json.JSONDecodeError:
                pass
            buffer = ""

    spans.reverse()
    return spans


def test_lazy_singleton_init(capfd):
    @observe()
    def trace_function():
        pass

    # Initialize the SDK after the decorator is defined
    verse.init(
        app_name="test_lazy_singleton_init",
        exporters=[verse.exporters.console()],
    )

    trace_function()
    out, _ = capfd.readouterr()
    spans = parse_spans(out)

    assert len(spans) == 1
    assert spans[0]["name"] == "trace_function"



from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import SpanProcessor
from opentelemetry.sdk.trace.export import (
    SimpleSpanProcessor,
)

from ...wrappers import FileSpanExporter, ScopeFilterExporter
from ..base import Exporter
from .types import FileConfig


class FileExporter(Exporter):
    config: FileConfig

    def __init__(self, config: FileConfig):
        self.config = config

    def create_span_processor(self, resource: Resource) -> SpanProcessor:
        """Create a span processor for the file exporter w/ scope filtering."""
        file_span_exporter = FileSpanExporter(self.config["folder_path"])
        exporter = ScopeFilterExporter(
            file_span_exporter,
            self.config,
        )

        return SimpleSpanProcessor(exporter)



import inspect
import json
from collections.abc import Callable
from typing import Any

from litellm import Message, acompletion, completion


def convert_functions_to_tools(functions: list[Callable]):
    tools = []

    for func in functions:
        params = {}
        required = []

        for param_name, param in inspect.signature(func).parameters.items():
            param_type = {
                bool: "boolean",
                dict: "object",
                int: "integer",
                list: "array",
            }.get(param.annotation, "string")

            params[param_name] = {
                "type": param_type,
                "description": f"Parameter {param_name}",
            }
            required.append(param_name)

        tools.append(
            {
                "function": {
                    "description": func.__doc__ or f"Function {func.__name__}",
                    "name": func.__name__,
                    "parameters": {
                        "properties": params,
                        "required": required,
                        "type": "object",
                    },
                },
                "type": "function",
            }
        )

    return tools


def create_function_map(functions: list[Callable]):
    return {func.__name__: func for func in functions}


def run_completion(
    messages: list[Message],
    funcs: list[Callable],
):
    function_map = create_function_map(funcs)
    iteration = 0
    max_iterations = 10
    tools = convert_functions_to_tools(funcs)

    while iteration < max_iterations:
        response = completion(
            model="openai/gpt-4o",
            messages=messages,
            tools=tools,
            tool_choice="auto",
        )

        message = response.choices[0].message
        tool_calls = message.tool_calls

        if tool_calls:
            messages.append(
                {
                    "content": message.content,
                    "role": "assistant",
                    "tool_calls": tool_calls,
                }
            )
            for tool_call in tool_calls:
                result = run_tool(tool_call, function_map)
                if result:
                    messages.append(result)
        else:
            break
        iteration += 1
    return response


async def run_completion_async(
    messages: list[Message],
    funcs: list[Callable],
):
    function_map = create_function_map(funcs)
    iteration = 0
    max_iterations = 10
    tools = convert_functions_to_tools(funcs)

    while iteration < max_iterations:
        response = await acompletion(
            model="openai/gpt-4o",  # note: account blocked from streaming newer models
            messages=messages,
            tools=tools,
            stream=True,
        )

        tool_calls_made = False
        tool_buffers = {}
        content_buffer = ""

        async for chunk in response:
            choice = chunk["choices"][0]
            delta = choice["delta"]

            if delta.get("content"):
                content_buffer += delta["content"]
                yield delta["content"]

            if delta.get("tool_calls"):
                tool_calls_made = True

                for tool_call in delta["tool_calls"]:
                    args_part = tool_call["function"].get("arguments", "")
                    call_id = tool_call["id"]
                    function_name = tool_call["function"].get("name", "")

                    if call_id is None and function_name is None:
                        if tool_buffers:
                            last_call_id = list(tool_buffers.keys())[-1]
                            tool_buffers[last_call_id]["function"]["arguments"] += (
                                args_part
                            )

                    else:
                        if call_id not in tool_buffers:
                            tool_buffers[call_id] = {
                                "id": call_id,
                                "function": {
                                    "name": function_name,
                                    "arguments": "",
                                },
                            }

                        tool_buffers[call_id]["function"]["arguments"] += args_part

        if content_buffer.strip():
            break

        if tool_calls_made and tool_buffers:
            tool_calls_for_message = []
            tool_results = []

            for call_id, tool_call_data in tool_buffers.items():
                if call_id is None:
                    continue

                try:
                    _ = json.loads(tool_call_data["function"]["arguments"])

                    # note: need to recreate the tool call object for run_tool
                    tool_call_obj = type(
                        "ToolCall",
                        (),
                        {
                            "id": call_id,
                            "function": type(
                                "Function",
                                (),
                                {
                                    "name": tool_call_data["function"]["name"],
                                    "arguments": tool_call_data["function"][
                                        "arguments"
                                    ],
                                },
                            )(),
                        },
                    )()

                    tool_result = run_tool(tool_call_obj, function_map)

                    if tool_result:
                        tool_results.append(tool_result)

                        tool_calls_for_message.append(
                            {
                                "id": call_id,
                                "function": {
                                    "arguments": tool_call_data["function"][
                                        "arguments"
                                    ],
                                    "name": tool_call_data["function"]["name"],
                                },
                                "type": "function",
                            }
                        )

                except json.JSONDecodeError:
                    continue
                except Exception:
                    continue

            if tool_calls_for_message:
                assistant_msg = {
                    "content": content_buffer,
                    "role": "assistant",
                    "tool_calls": tool_calls_for_message,
                }
                messages.append(assistant_msg)

                for tool_result in tool_results:
                    messages.append(tool_result)

        if not tool_calls_made and not content_buffer.strip():
            break

        iteration += 1


def run_tool(tool: Any, function_map: dict[str, Callable]):
    function_name = tool.function.name
    function_args = json.loads(tool.function.arguments)

    if function_name in function_map:
        try:
            result = function_map[function_name](**function_args)
            return {
                "content": json.dumps(result),
                "role": "tool",
                "tool_call_id": tool.id,
                "id": tool.id,
            }
        except Exception:
            return None



from __future__ import annotations

import logging
from contextlib import contextmanager
from typing import Any

from cuid2 import Cuid
from opentelemetry import trace
from opentelemetry.sdk.trace import Span, Tracer

from .contexts import (
    ContextMetadata,
    EventMetadata,
    GenerationContext,
    GenerationMessage,
    GenerationUsage,
    ObservationType,
    Score,
    SpanContext,
    TraceContext,
)
from .utils import apply_value


class VerseSDK:
    """Core context manager factory"""

    tracer: trace.Tracer

    def __init__(self, tracer: Tracer):
        self._set_tracer(tracer)
        self._validate_tracer_provider()

    def create_generation(
        self,
        name: str,
        *,
        input: Any | None = None,
        messages: list[GenerationMessage] | None = None,
        model: str | None = None,
        output: Any | None = None,
        span: Span | None = None,
        usage: GenerationUsage | None = None,
        vendor: str | None = None,
        **attrs,
    ) -> GenerationContext:
        """
        Create or configure a generation span (LLM operation).

        Args:
            name: Name of the generation span
            messages: Optional list of messages
            model: Model name/identifier
            metadata: Optional metadata dictionary
            usage: Optional usage dictionary
            **attrs: Additional attributes to set on the span

        Returns:
            GenerationContext: A context object for the generation span
        """
        gen_ctx = GenerationContext(span or self._tracer.start_span(name))
        apply_value(gen_ctx, "input", input)
        apply_value(gen_ctx, "messages", messages)
        apply_value(gen_ctx, "model", model)
        apply_value(gen_ctx, "output", output)
        apply_value(gen_ctx, "usage", usage)
        apply_value(gen_ctx, "vendor", vendor)
        return gen_ctx.set_attributes(**attrs)

    def create_span(
        self,
        name: str,
        *,
        input: Any | None = None,
        level: str | None = None,
        metadata: ContextMetadata | None = None,
        op: str | None = None,
        output: Any | None = None,
        span: Span | None = None,
        status_message: str | None = None,
        **attrs,
    ):
        """
        Create or configure a new span (sub-operation).

        Args:
            name: Name of the span
            input: Optional input for the span
            level: Optional level for the span
            metadata: Optional metadata dictionary
            op: Optional operation for the span
            output: Optional output for the span
            status_message: Optional status message for the span
            **attrs: Additional attributes to set on the span

        Returns:
            SpanContext: A context object for the span
        """
        span_ctx = SpanContext(span or self._tracer.start_span(name))
        apply_value(span_ctx, "input", input)
        apply_value(span_ctx, "level", level)
        apply_value(span_ctx, "metadata", metadata)
        apply_value(span_ctx, "operation", op)
        apply_value(span_ctx, "output", output)
        apply_value(span_ctx, "status_message", status_message)
        return span_ctx.set_attributes(**attrs)

    def create_trace(
        self,
        name: str,
        *,
        metadata: ContextMetadata | None = None,
        scope: str | None = None,
        session_id: str | None = None,
        span: Span | None = None,
        tags: list[str] | None = None,
        user_id: str | None = None,
        **attrs,
    ):
        """
        Create or configure a new trace span (top-level operation).

        Args:
            metadata: Optional metadata dictionary
            name: Name of the trace
            scope: Optional scope for the trace
            session_id: Optional session identifier
            tags: Optional list of tags
            user_id: Optional user identifier
            **attrs: Additional attributes to set on the span

        Returns:
            TraceContext: A context object for the trace span
        """
        if not session_id:
            logging.warning("No session ID provided, generating a new one")
            session_id = Cuid().generate()

        trace_ctx = TraceContext(span or self._tracer.start_span(name))
        apply_value(trace_ctx, "metadata", metadata)
        apply_value(trace_ctx, "scope", scope)
        apply_value(trace_ctx, "session", session_id)
        apply_value(trace_ctx, "tags", tags)
        apply_value(trace_ctx, "user", user_id)
        return trace_ctx.set_attributes(**attrs)

    @contextmanager
    def generation(
        self,
        name: str,
        *,
        input: Any | None = None,
        messages: list[GenerationMessage] | None = None,
        model: str | None = None,
        output: Any | None = None,
        usage: GenerationUsage | None = None,
        vendor: str | None = None,
        **attrs,
    ):
        """
        Create a generation span as a context manager (LLM operation).

        Args:
            name: Name of the generation span
            messages: Optional list of messages
            model: Model name/identifier
            metadata: Optional metadata dictionary
            usage: Optional usage dictionary
            **attrs: Additional attributes to set on the span

        Yields:
            GenerationContext: A context object for the generation span
        """
        with self._tracer.start_as_current_span(name) as span:
            yield self.create_generation(
                name,
                input=input,
                messages=messages,
                model=model,
                output=output,
                span=span,
                usage=usage,
                vendor=vendor,
                **attrs,
            )

    @contextmanager
    def span(
        self,
        name: str,
        *,
        input: Any | None = None,
        level: str | None = None,
        metadata: ContextMetadata | None = None,
        op: str | None = None,
        output: Any | None = None,
        status_message: str | None = None,
        **attrs,
    ):
        """
        Create a new span as a context manager (sub-operation).

        Args:
            name: Name of the span
            input: Optional input for the span
            level: Optional level for the span
            metadata: Optional metadata dictionary
            op: Optional operation for the span
            output: Optional output for the span
            status_message: Optional status message for the span
            **attrs: Additional attributes to set on the span

        Yields:
            SpanContext: A context object for the span
        """
        with self._tracer.start_as_current_span(name) as span:
            yield self.create_span(
                name,
                input=input,
                level=level,
                metadata=metadata,
                op=op,
                output=output,
                status_message=status_message,
                span=span,
                **attrs,
            )

    @contextmanager
    def trace(
        self,
        name: str,
        *,
        metadata: ContextMetadata | None = None,
        scope: str | None = None,
        session_id: str | None = None,
        tags: list[str] | None = None,
        user_id: str | None = None,
        **attrs,
    ):
        """
        Create a new trace span as a context manager (top-level operation).

        Args:
            metadata: Optional metadata dictionary
            name: Name of the trace
            scope: Optional scope for the trace
            session_id: Optional session identifier
            tags: Optional list of tags
            user_id: Optional user identifier
            **attrs: Additional attributes to set on the span

        Yields:
            TraceContext: A context object for the trace span
        """
        with self._tracer.start_as_current_span(name) as span:
            yield self.create_trace(
                name,
                metadata=metadata,
                scope=scope,
                session_id=session_id,
                span=span,
                tags=tags,
                user_id=user_id,
                **attrs,
            )

    def event(
        self,
        name: str,
        *,
        level: str = "info",
        metadata: EventMetadata | None = None,
        span: Span | None = None,
        **attrs,
    ) -> None:
        """
        Add an event to the current span.

        Args:
            name: Event name
            level: Event level (e.g., "info", "warning", "error")
            metadata: Optional metadata dictionary
            span: Optional span to add the event to
            **attrs: Additional event attributes

        Returns:
            None
        """
        try:
            SpanContext(span or trace.get_current_span()).event(
                name, level=level, metadata=metadata, **attrs
            )
        except Exception as e:
            logging.warning("Error adding event to span", exc_info=e)

    def score(
        self,
        score: Score,
        target: ObservationType,
    ) -> None:
        """
        Add a score to the current observation or trace.

        Args:
            name: Score name/metric
            score: Score dictionary
            target: Whether to score the observation or entire trace
        """
        try:
            span = trace.get_current_span()
            if target == "observation":
                SpanContext(span).score(score)
            else:
                TraceContext(span).score(score)
        except Exception as e:
            logging.warning("Error adding score to span", exc_info=e)

    def _set_tracer(self, tracer: Tracer | None) -> None:
        """Check null check when setting tracer instance"""
        if not tracer:
            raise ValueError("Tracer not initialized")
        self._tracer = tracer

    def _validate_tracer_provider(self) -> None:
        """Validate tracer provider"""
        provider = None

        try:
            provider = trace.get_tracer_provider()
        except Exception as e:
            raise ValueError("Tracer provider unavailable") from e

        if not provider or isinstance(provider, trace.NoOpTracerProvider):
            raise ValueError("Invalid tracer provider given")



from __future__ import annotations

from unittest.mock import Mock

from verse_sdk.exporters.types import ExporterConfig
from verse_sdk.wrappers.scope_filter import ScopeFilterExporter


class TestScopeFilterExporter:
    def setup_method(self):
        self.mock_exporter = Mock()
        self.mock_span_context = Mock()
        self.mock_span_context.trace_id = 12345

    def _create_mock_span(self, trace_id=12345, scope=None, attributes=None):
        span = Mock()
        span.get_span_context.return_value.trace_id = trace_id
        span.attributes = attributes or {}
        if scope is not None:
            span.attributes["session.scope"] = scope
        return span

    def test_initialization_with_scopes(self):
        config: ExporterConfig = {"scopes": ["user", "admin"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        assert exporter._exporter == self.mock_exporter
        assert exporter._allowed_scopes == ["user", "admin"]

    def test_initialization_with_empty_scopes(self):
        config: ExporterConfig = {"scopes": []}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        assert exporter._exporter == self.mock_exporter
        assert exporter._allowed_scopes == []

    def test_initialization_with_none_scopes(self):
        config: ExporterConfig = {"scopes": None}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        assert exporter._exporter == self.mock_exporter
        assert exporter._allowed_scopes is None

    def test_initialization_without_scopes_key(self):
        config: ExporterConfig = {}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        assert exporter._exporter == self.mock_exporter
        assert exporter._allowed_scopes == []

    def test_export_with_no_scopes_configured(self):
        config: ExporterConfig = {"scopes": []}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        spans = [
            self._create_mock_span(trace_id=1, scope="user"),
            self._create_mock_span(trace_id=2, scope="admin"),
            self._create_mock_span(trace_id=3, scope="guest"),
        ]

        exporter.export(spans)
        self.mock_exporter.export.assert_called_once_with(spans)

    def test_export_with_none_scopes_configured(self):
        config: ExporterConfig = {"scopes": None}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        spans = [
            self._create_mock_span(trace_id=1, scope="user"),
            self._create_mock_span(trace_id=2, scope="admin"),
        ]

        exporter.export(spans)
        self.mock_exporter.export.assert_called_once_with(spans)

    def test_export_without_scopes_key(self):
        config: ExporterConfig = {}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        spans = [
            self._create_mock_span(trace_id=1, scope="user"),
            self._create_mock_span(trace_id=2, scope="admin"),
        ]

        exporter.export(spans)
        self.mock_exporter.export.assert_called_once_with(spans)

    def test_export_filters_spans_by_allowed_scopes(self):
        config: ExporterConfig = {"scopes": ["user", "admin"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        user_span = self._create_mock_span(trace_id=1, scope="user")
        admin_span = self._create_mock_span(trace_id=2, scope="admin")
        guest_span = self._create_mock_span(trace_id=3, scope="guest")
        no_scope_span = self._create_mock_span(trace_id=4, scope=None)

        spans = [user_span, admin_span, guest_span, no_scope_span]
        exporter.export(spans)
        expected_filtered_spans = [user_span, admin_span]
        self.mock_exporter.export.assert_called_once_with(expected_filtered_spans)

    def test_export_filters_spans_by_single_allowed_scope(self):
        config: ExporterConfig = {"scopes": ["admin"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        user_span = self._create_mock_span(trace_id=1, scope="user")
        admin_span = self._create_mock_span(trace_id=2, scope="admin")
        guest_span = self._create_mock_span(trace_id=3, scope="guest")

        spans = [user_span, admin_span, guest_span]
        exporter.export(spans)

        expected_filtered_spans = [admin_span]
        self.mock_exporter.export.assert_called_once_with(expected_filtered_spans)

    def test_export_includes_all_spans_from_trace_with_allowed_scope(self):
        config: ExporterConfig = {"scopes": ["user"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        user_span1 = self._create_mock_span(trace_id=1, scope="user")
        user_span2 = self._create_mock_span(trace_id=1, scope=None)
        admin_span = self._create_mock_span(trace_id=2, scope="admin")

        spans = [user_span1, user_span2, admin_span]
        exporter.export(spans)

        expected_filtered_spans = [user_span1, user_span2]
        self.mock_exporter.export.assert_called_once_with(expected_filtered_spans)

    def test_export_excludes_all_spans_from_trace_without_allowed_scope(self):
        config: ExporterConfig = {"scopes": ["user"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        admin_span1 = self._create_mock_span(trace_id=1, scope="admin")
        admin_span2 = self._create_mock_span(trace_id=1, scope=None)
        guest_span = self._create_mock_span(trace_id=2, scope="guest")

        spans = [admin_span1, admin_span2, guest_span]

        exporter.export(spans)

        expected_filtered_spans = []
        self.mock_exporter.export.assert_called_once_with(expected_filtered_spans)

    def test_export_with_empty_spans_list(self):
        config: ExporterConfig = {"scopes": ["user"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)
        exporter.export([])

        self.mock_exporter.export.assert_called_once_with([])

    def test_shutdown_delegates_to_exporter(self):
        config: ExporterConfig = {"scopes": ["user"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        exporter.shutdown()

        self.mock_exporter.shutdown.assert_called_once()

    def test_force_flush_delegates_to_exporter(self):
        config: ExporterConfig = {"scopes": ["user"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        exporter.force_flush(5000)

        self.mock_exporter.force_flush.assert_called_once_with(5000)

    def test_force_flush_with_no_timeout(self):
        config: ExporterConfig = {"scopes": ["user"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        exporter.force_flush()

        self.mock_exporter.force_flush.assert_called_once_with(None)

    def test_export_with_mixed_scope_scenarios(self):
        config: ExporterConfig = {"scopes": ["user", "guest"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        spans = [
            self._create_mock_span(trace_id=1, scope="user"),
            self._create_mock_span(trace_id=1, scope=None),
            self._create_mock_span(trace_id=2, scope="admin"),
            self._create_mock_span(trace_id=2, scope=None),
            self._create_mock_span(trace_id=3, scope="guest"),
            self._create_mock_span(trace_id=4, scope=None),
        ]

        exporter.export(spans)
        self.mock_exporter.export.assert_called_once()
        call_args = self.mock_exporter.export.call_args[0][0]
        assert len(call_args) == 3

    def test_export_with_span_attributes_access(self):
        config: ExporterConfig = {"scopes": ["user"]}
        exporter = ScopeFilterExporter(self.mock_exporter, config)

        span = Mock()
        span.get_span_context.return_value.trace_id = 1
        span.attributes = {"session.scope": "user", "other.attr": "value"}

        spans = [span]

        exporter.export(spans)

        assert "session.scope" in span.attributes
        self.mock_exporter.export.assert_called_once()



from .contexts import (
    get_current_generation_context,
    get_current_span_context,
    get_current_trace_context,
)
from .decorators import create_decorator
from .exporters import (
    FileConfig,
    LangfuseConfig,
    OtelConfig,
    VerseConfig,
)
from .sdk_builder import VerseSDKBuilder
from .spec import OtelGenAISpecCheck

verse = VerseSDKBuilder()
exporters = verse.exporters
observe = create_decorator(verse)
shutdown = verse.shutdown


__all__ = [
    "FileConfig",
    "LangfuseConfig",
    "OtelConfig",
    "OtelGenAISpecCheck",
    "VerseConfig",
    "exporters",
    "get_current_generation_context",
    "get_current_span_context",
    "get_current_trace_context",
    "observe",
    "shutdown",
    "verse",
]



import logging
from functools import wraps

from .base import Exporter
from .console import ConsoleExporter
from .file import FileConfig, FileExporter
from .langfuse import LangfuseConfig, LangfuseExporter
from .otel import OtelConfig, OTLPExporter
from .types import ExporterConfig
from .verse import VerseConfig, VerseExporter


class ExportersRegistry:
    @staticmethod
    def safe_named_exporter(fn):
        """Create an error boundary around each exporter factory function."""

        @wraps(fn)
        def create_exporter(*args, **kwargs) -> Exporter | None:
            try:
                return fn(*args, **kwargs)
            except Exception as e:
                logging.warning("Error creating exporter `%s`", fn.__name__, exc_info=e)
                return None

        return create_exporter

    @safe_named_exporter
    def console(self, config: ExporterConfig = None) -> ConsoleExporter:
        """Create a console exporter."""
        return ConsoleExporter(config or {})

    @safe_named_exporter
    def file(self, config: FileConfig = None) -> FileExporter:
        """Create a file exporter."""
        return FileExporter(config or {})

    @safe_named_exporter
    def langfuse(self, config: LangfuseConfig = None) -> LangfuseExporter:
        """Create a langfuse exporter."""
        return LangfuseExporter(config or {})

    @safe_named_exporter
    def otel(self, config: OtelConfig = None) -> OTLPExporter:
        """Create a otel exporter."""
        return OTLPExporter(config or {})

    @safe_named_exporter
    def verse(self, config: VerseConfig = None) -> VerseExporter:
        """Create a verse exporter."""
        return VerseExporter(config or {})


__all__ = [
    "Exporter",
    "ExportersRegistry",
    "FileConfig",
    "LangfuseConfig",
    "OtelConfig",
    "VerseConfig",
]



import pytest

from tests.e2e.runner import E2ERunnerAsync


@pytest.mark.e2e
@pytest.mark.parametrize(
    "name",
    [
        "anthropic",
        # No paid account to run in CI/CD
        # "google",
        "langchain",
        "litellm",
        "pydantic_ai",
    ],
)
def test_agent_integration_with_streaming(name: str):
    E2ERunnerAsync(name).run()



def as_traces_endpoint(host: str | None, api_key: str | None = None) -> str:
    """Convert a host to a traces endpoint."""

    traces_endpoint = ""

    if host:
        traces_endpoint = host if host.endswith("traces") else f"{host}/v1/traces"

    if api_key:
        return f"{traces_endpoint}?api_key={api_key}"

    return traces_endpoint
